unit UDpiControls;

{$mode objfpc}{$H+}

interface

uses
  {$IFDEF WINDOWS}Windows, {$ENDIF}Classes, SysUtils, LCLProc, LResources, Forms, FormEditingIntf, ProjectIntf,
  Controls, StdCtrls, fgl, Graphics, ComCtrls, ExtCtrls, LCLType, GraphType,
  Types, CustApp, LMessages, LCLIntf, Menus, Math, UPixelPointer2;

type
  TMessageEvent = procedure (var TheMessage : TLMessage) of object;

  { TFormEx }

  TFormEx = class(TForm)
  private
    FOnMessage: TMessageEvent;
  protected
    procedure WndProc(var TheMessage : TLMessage); override;
    property OnMessage: TMessageEvent read FOnMessage write FOnMessage;
  end;

  { TGraphicControlEx }

  TGraphicControlEx = class(TGraphicControl)
  public
    property OnPaint;
    procedure Paint; override;
  end;

  { TControlEx }

  TControlEx = class(TControl)
  public
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnMouseWheel;
    property OnMouseLeave;
    property OnMouseEnter;
    property ParentFont;
  end;

  TWinControlEx = class(TWinControl)
  public
    property BorderStyle;
  end;

  { TDpiFormFileDesc }

  TDpiFormFileDesc = class(TFileDescPascalUnitWithResource)
  public
    constructor Create; override;
    function GetInterfaceUsesSection: string; override;
    function GetLocalizedName: string; override;
    function GetLocalizedDescription: string; override;
  end;

  { TDpiFont }

  TDpiFont = class(TPersistent)
  private
    FNativeFont: TFont;
    FNativeFontFree: Boolean;
    FOnChange: TNotifyEvent;
    FSize: Integer;
    FHeight: Integer;
    FPixelsPerInch: Integer;
    FColor: TColor;
    function GetCharSet: TFontCharSet;
    function GetColor: TColor;
    function GetHeight: Integer;
    function GetName: string;
    function GetPixelsPerInch: Integer;
    function GetSize: Integer;
    function GetStyle: TFontStyles;
    function IsNameStored: Boolean;
    procedure SetCharSet(AValue: TFontCharSet);
    procedure SetColor(AValue: TColor);
    procedure SetHeight(AValue: Integer);
    procedure SetName(AValue: string);
    procedure SetNativeFont(AValue: TFont);
    procedure SetOnChange(AValue: TNotifyEvent);
    procedure SetPixelsPerInch(AValue: Integer);
    procedure SetSize(AValue: Integer);
    procedure DoChange;
    procedure SetStyle(AValue: TFontStyles);
    procedure UpdateFont;
  protected
    procedure ScreenChanged;
    function GetNativeFont: TFont; virtual;
  public
    property NativeFont: TFont read FNativeFont write SetNativeFont;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
  published
    property CharSet: TFontCharSet read GetCharSet write SetCharSet default DEFAULT_CHARSET;
    property Color: TColor read GetColor write SetColor;
    property Name: string read GetName write SetName stored IsNameStored;
    property Style: TFontStyles read GetStyle write SetStyle default [];
    property Size: Integer read GetSize write SetSize stored false;
    property PixelsPerInch: Integer read GetPixelsPerInch write SetPixelsPerInch;
    property Height: Integer read GetHeight write SetHeight default 0;
    property OnChange: TNotifyEvent read FOnChange write SetOnChange;
  end;

  { TDpiSizeConstraints }

  TDpiSizeConstraints = class(TPersistent)
  private
    FMaxHeight: TConstraintSize;
    FMaxWidth: TConstraintSize;
    FMinHeight: TConstraintSize;
    FMinWidth: TConstraintSize;
    procedure SetMaxHeight(AValue: TConstraintSize);
    procedure SetMaxWidth(AValue: TConstraintSize);
    procedure SetMinHeight(AValue: TConstraintSize);
    procedure SetMinWidth(AValue: TConstraintSize);
  published
    property MaxHeight: TConstraintSize read FMaxHeight write SetMaxHeight default 0;
    property MaxWidth: TConstraintSize read FMaxWidth write SetMaxWidth default 0;
    property MinHeight: TConstraintSize read FMinHeight write SetMinHeight default 0;
    property MinWidth: TConstraintSize read FMinWidth write SetMinWidth default 0;
  end;


  TDpiWinControl = class;
  TDpiControlBorderSpacing = class;

  { TDpiControl }

  TDpiControl = class(TComponent)
  private
    FBorderSpacing: TDpiControlBorderSpacing;
    FConstraints: TDpiSizeConstraints;
    FFont: TDpiFont;
    FHeight: Integer;
    FLeft: Integer;
    FOnChangeBounds: TNotifyEvent;
    FOnMouseUp: TMouseEvent;
    FOnMouseDown: TMouseEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnMouseWheel: TMouseWheelEvent;
    FOnResize: TNotifyEvent;
    FTop: Integer;
    FWidth: Integer;
    FParent: TDpiWinControl;
    function GetAlign: TAlign;
    function GetAnchors: TAnchors;
    function GetBoundsRect: TRect;
    function GetClientHeight: Integer;
    function GetClientWidth: Integer;
    function GetColor: TColor;
    function GetCursor: TCursor;
    function GetEnabled: Boolean;
    function GetHint: string;
    function GetOnClick: TNotifyEvent;
    function GetParentFont: Boolean;
    function GetShowHint: Boolean;
    function GetVisible: Boolean;
    function IsAnchorsStored: Boolean;
    procedure SetAlign(AValue: TAlign);
    procedure SetAnchors(AValue: TAnchors);
    procedure SetBorderSpacing(AValue: TDpiControlBorderSpacing);
    procedure SetBoundsRect(AValue: TRect);
    procedure SetClientHeight(AValue: Integer);
    procedure SetClientWidth(AValue: Integer);
    procedure SetColor(AValue: TColor);
    procedure SetCursor(AValue: TCursor);
    procedure SetEnabled(AValue: Boolean);
    procedure SetFont(AValue: TDpiFont);
    procedure SetHint(AValue: string);
    procedure SetOnChangeBounds(AValue: TNotifyEvent);
    procedure SetOnClick(AValue: TNotifyEvent);
    procedure SetOnResize(AValue: TNotifyEvent);
    procedure SetParentFont(AValue: Boolean);
    procedure SetShowHint(AValue: Boolean);
    procedure NativeFormResize(Sender: TObject);
    procedure NativeChangeBounds(Sender: TObject);
    procedure DoFormResize;
    procedure DoChangeBounds;
    procedure MouseDownHandler(Sender: TObject; Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); virtual;
    procedure MouseUpHandler(Sender: TObject; Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); virtual;
    procedure MouseMoveHandler(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
    procedure MouseWheelHandler(Sender: TObject; Shift: TShiftState;
         WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean); virtual;
    procedure MouseLeaveHandler(Sender: TObject); virtual;
    procedure MouseEnterHandler(Sender: TObject); virtual;
  protected
    procedure DoBorderSpacingChange(Sender: TObject; InnerSpaceChanged: Boolean); virtual;
    function GetText: TCaption; virtual;
    procedure SetText(AValue: TCaption); virtual;
    procedure UpdateBounds; virtual;
    procedure FontChanged(Sender: TObject); virtual;
    function GetCaption: string; virtual;
    procedure SetParent(AValue: TDpiWinControl); virtual;
    procedure SetCaption(AValue: string); virtual;
    procedure SetHeight(AValue: Integer); virtual;
    procedure SetLeft(AValue: Integer); virtual;
    procedure SetTop(AValue: Integer); virtual;
    procedure SetVisible(AValue: Boolean); virtual;
    procedure SetWidth(AValue: Integer); virtual;
    function GetNativeControl: TControl; virtual;
    procedure UpdateNativeControl; virtual;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); virtual;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); virtual;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); virtual;
    procedure MouseLeave; virtual;
    procedure MouseEnter; virtual;
    property Text: TCaption read GetText write SetText;
    property ParentFont: Boolean read GetParentFont write SetParentFont default True;
  public
    function ScreenToClient(const APoint: TPoint): TPoint; virtual;
    function ClientToScreen(const APoint: TPoint): TPoint; virtual;
    procedure AddHandlerOnVisibleChanged(const OnVisibleChangedEvent: TNotifyEvent;
                                         AsFirst: boolean = false);
    procedure RemoveHandlerOnVisibleChanged(const OnVisibleChangedEvent: TNotifyEvent);
    procedure ScreenChanged; virtual;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure Show;
    procedure Hide;
    procedure Invalidate;
    procedure Repaint;
    procedure Update;
    function IsParentOf(AControl: TDpiControl): boolean; virtual;
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    property Parent: TDpiWinControl read FParent write SetParent;
    property BoundsRect: TRect read GetBoundsRect write SetBoundsRect;
    property Visible: Boolean read GetVisible write SetVisible;
    property Anchors: TAnchors read GetAnchors write SetAnchors stored IsAnchorsStored default [akLeft, akTop];
    property BorderSpacing: TDpiControlBorderSpacing read FBorderSpacing write SetBorderSpacing;
  published
    property ClientHeight: Integer read GetClientHeight write SetClientHeight;
    property ClientWidth: Integer read GetClientWidth write SetClientWidth;
    property Hint: string read GetHint write SetHint;
    property Top: Integer read FTop write SetTop;
    property Left: Integer read FLeft write SetLeft;
    property Width: Integer read FWidth write SetWidth;
    property Height: Integer read FHeight write SetHeight;
    property Caption: string read GetCaption write SetCaption;
    property Enabled: Boolean read GetEnabled write SetEnabled;
    property ShowHint: Boolean read GetShowHint write SetShowHint;
    property Font: TDpiFont read FFont write SetFont;
    property Align: TAlign read GetAlign write SetAlign;
    property Color: TColor read GetColor write SetColor;
    property Constraints: TDpiSizeConstraints read FConstraints write FConstraints;
    property Cursor: TCursor read GetCursor write SetCursor default crDefault;
    property OnResize: TNotifyEvent read FOnResize write SetOnResize;
    property OnChangeBounds: TNotifyEvent read FOnChangeBounds write SetOnChangeBounds;
    property OnClick: TNotifyEvent read GetOnClick write SetOnClick;
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnMouseWheel: TMouseWheelEvent read FOnMouseWheel write FOnMouseWheel;
  end;

  TDpiControls = specialize TFPGObjectList<TDpiControl>;

  { TDpiControlBorderSpacing }

  TDpiControlBorderSpacing = class(TPersistent)
  private
    FAround: TSpacingSize;
    FBottom: TSpacingSize;
    FLeft: TSpacingSize;
    FOnChange: TNotifyEvent;
    FRight: TSpacingSize;
    FTop: TSpacingSize;
    FControl: TDpiControl;
    FDefault: PControlBorderSpacingDefault;
    function IsAroundStored: Boolean;
    function IsBottomStored: Boolean;
    function IsLeftStored: Boolean;
    function IsRightStored: Boolean;
    function IsTopStored: Boolean;
    procedure SetAround(AValue: TSpacingSize);
    procedure SetBottom(AValue: TSpacingSize);
    procedure SetLeft(AValue: TSpacingSize);
    procedure SetRight(AValue: TSpacingSize);
    procedure SetTop(AValue: TSpacingSize);
    procedure Change(InnerSpaceChanged: Boolean); virtual;
  public
    constructor Create(OwnerControl: TDpiControl; ADefault: PControlBorderSpacingDefault = nil);
  published
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Around: TSpacingSize read FAround write SetAround stored IsAroundStored;
    property Left: TSpacingSize read FLeft write SetLeft stored IsLeftStored;
    property Top: TSpacingSize read FTop write SetTop stored IsTopStored;
    property Right: TSpacingSize read FRight write SetRight stored IsRightStored;
    property Bottom: TSpacingSize read FBottom write SetBottom stored IsBottomStored;
  end;

  { TDpiGraphic }

  TDpiGraphic = class(TPersistent)
  protected
    FDpi: Integer;
    function GetNativeGraphic: TGraphic; virtual;
    function GetWidth: Integer; virtual; abstract;
    function GetHeight: Integer; virtual; abstract;
    procedure SetWidth(Value: Integer); virtual; abstract;
    procedure SetHeight(Value: Integer); virtual; abstract;
    procedure ScreenChanged; virtual;
    procedure SetDpi(AValue: Integer); virtual;
    function GetDpi: Integer; virtual;
  public
    NativeGraphicClass: TGraphicClass;
    constructor Create; virtual;
    procedure LoadFromFile(const Filename: string); virtual;
    procedure SaveToFile(const Filename: string); virtual;
    property Width: Integer read GetWidth write SetWidth;
    property Height: Integer read GetHeight write SetHeight;
    property Dpi: Integer read GetDpi write SetDpi;
  end;

  { TDpiRasterImage }

  TDpiRasterImage = class(TDpiGraphic)
  private
    function GetRawImage: TRawImage;
  protected
    function GetNativeGraphic: TGraphic; override;
    function GetNativeRasterImage: TRasterImage; virtual;
  public
    property RawImage: TRawImage read GetRawImage;
  end;

  { TDpiCanvas }

  TDpiCanvas = class
  private
    FFont: TDpiFont;
    FFontFree: Boolean;
    FNativeCanvas: TCanvas;
    FNativeCanvasFree: Boolean;
    function GetBrush: TBrush;
    function GetHandle: HDC;
    function GetHeight: Integer;
    function GetPen: TPen;
    function GetPixel(X, Y: Integer): TColor;
    function GetWidth: Integer;
    procedure SetBrush(AValue: TBrush);
    procedure SetFont(AValue: TDpiFont);
    procedure SetHandle(AValue: HDC);
    procedure SetPen(AValue: TPen);
    procedure SetPixel(X, Y: Integer; AValue: TColor);
    procedure SetNativeCanvas(AValue: TCanvas);
  protected
    function GetNativeCanvas: TCanvas; virtual;
  public
    property NativeCanvas: TCanvas read FNativeCanvas write SetNativeCanvas;
    procedure StretchDraw(const DestRect: TRect; SrcGraphic: TDpiGraphic); virtual;
    procedure FrameRect(Rect: TRect);
    procedure Rectangle(X1, Y1, X2, Y2: Integer); overload;
    procedure Rectangle(const ARect: TRect); overload;
    function TextWidth(Text: string): Integer;
    function TextHeight(Text: string): Integer;
    function TextExtent(Text: string): TSize;
    procedure TextOut(X, Y: Integer; Text: string);
    procedure TextRect(ARect: TRect; X, Y: Integer; Text: string);
    procedure MoveTo(X, Y: Integer);
    procedure LineTo(X, Y: Integer);
    procedure FillRect(ARect: TRect);
    procedure FillRect(X1, Y1, X2, Y2: Integer);
    procedure Draw(X, Y: Integer; Source: TDpiGraphic);
    procedure CopyRect(Dest: TRect; SrcCanvas: TDpiCanvas; Source: TRect);
    constructor Create;
    destructor Destroy; override;
    property Handle: HDC read GetHandle write SetHandle;
    property Pixels[X, Y: Integer]: TColor read GetPixel write SetPixel;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
  published
    property Brush: TBrush read GetBrush write SetBrush;
    property Pen: TPen read GetPen write SetPen;
    property Font: TDpiFont read FFont write SetFont;
  end;

  { TDpiGraphicControl }

  TDpiGraphicControl = class(TDpiControl)
  private
    FOnPaint: TNotifyEvent;
    NativeGraphicControl: TGraphicControl;
    FCanvas: TDpiCanvas;
    function GetOnPaint: TNotifyEvent;
    procedure SetCanvas(AValue: TDpiCanvas);
    procedure PaintHandler(Sender: TObject);
    procedure SetOnPaint(AValue: TNotifyEvent);
  protected
    procedure Paint; virtual;
    function GetNativeControl: TControl; override;
    function GetNativeGraphicControl: TGraphicControl; virtual;
    procedure UpdateNativeControl; override;
    property OnPaint: TNotifyEvent read GetOnPaint write SetOnPaint;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Canvas: TDpiCanvas read FCanvas write SetCanvas;
  end;

  { TDpiWinControl }

  TDpiWinControl = class(TDpiControl)
  private
    function GetBorderStyle: TBorderStyle;
    function GetHandle: HWND;
    function GetOnKeyDown: TKeyEvent;
    function GetOnKeyPress: TKeyPressEvent;
    function GetOnKeyUp: TKeyEvent;
    function GetTabOrder: TTabOrder;
    function GetTabStop: Boolean;
    procedure SetBorderStyle(AValue: TBorderStyle);
    procedure SetHandle(AValue: HWND);
    procedure SetOnKeyDown(AValue: TKeyEvent);
    procedure SetOnKeyPress(AValue: TKeyPressEvent);
    procedure SetOnKeyUp(AValue: TKeyEvent);
    procedure SetTabOrder(AValue: TTabOrder);
    procedure SetTabStop(AValue: Boolean);
  protected
    function GetNativeControl: TControl; override;
    function GetNativeWinControl: TWinControl; virtual;
    property BorderStyle: TBorderStyle read GetBorderStyle write SetBorderStyle default bsNone;
  public
    Controls: TDpiControls;
    procedure ScreenChanged; override;
    function ControlCount: Integer;
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    property Handle: HWND read GetHandle write SetHandle;
  published
    property TabOrder: TTabOrder read GetTabOrder write SetTabOrder default -1;
    property TabStop: Boolean read GetTabStop write SetTabStop default False;
    property OnKeyDown: TKeyEvent read GetOnKeyDown write SetOnKeyDown;
    property OnKeyPress: TKeyPressEvent read GetOnKeyPress write SetOnKeyPress;
    property OnKeyUp: TKeyEvent read GetOnKeyUp write SetOnKeyUp;
  end;

  { TDpiCustomControl }

  TDpiCustomControl = class(TDpiWinControl)
  private
    FCanvas: TDpiCanvas;
    function GetCanvas: TDpiCanvas;
    function GetOnPaint: TNotifyEvent;
    function GetPixelsPerInch: Integer;
    procedure SetOnPaint(AValue: TNotifyEvent);
    procedure SetPixelsPerInch(AValue: Integer);
  protected
    function GetNativeWinControl: TWinControl; override;
    function GetNativeCustomControl: TCustomControl; virtual;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    property Canvas: TDpiCanvas read GetCanvas;
  published
    property PixelsPerInch: Integer read GetPixelsPerInch write SetPixelsPerInch stored False;
    property OnPaint: TNotifyEvent read GetOnPaint write SetOnPaint;
  end;

  { TDpiControlScrollBar }

  TDpiControlScrollBar = class(TPersistent)
  private
    function GetVisible: Boolean;
    procedure SetVisible(AValue: Boolean);
  published
    property Visible: Boolean read GetVisible write SetVisible;
  end;

  { TDpiScrollingWinControl }

  TDpiScrollingWinControl = class(TDpiCustomControl)
  private
    FHorzScrollBar: TDpiControlScrollBar;
    FVertScrollBar: TDpiControlScrollBar;
  protected
    function GetNativeCustomControl: TCustomControl; override;
    function GetNativeScrollingWinControl: TScrollingWinControl; virtual;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
  published
    property HorzScrollBar: TDpiControlScrollBar read FHorzScrollBar write FHorzScrollBar;
    property VertScrollBar: TDpiControlScrollBar read FVertScrollBar write FVertScrollBar;
  end;

  { TDpiForm }

  TDpiForm = class(TDpiScrollingWinControl)
  private
    FOnActivate: TNotifyEvent;
    FOnClose: TCloseEvent;
    FOnCloseQuery: TCloseQueryEvent;
    FOnDeactivate: TNotifyEvent;
    function GetBorderIcons: TBorderIcons;
    function GetBorderStyle: TFormBorderStyle;
    function GetDesignTimePPI: Integer;
    function GetFormState: TFormState;
    function GetFormStyle: TFormStyle;
    function GetKeyPreview: Boolean;
    function GetLCLVersion: string;
    function GetModalResult: TModalResult;
    function GetOnCloseQuery: TCloseQueryEvent;
    function GetOnCreate: TNotifyEvent;
    function GetOnDeactivate: TNotifyEvent;
    function GetOnDestroy: TNotifyEvent;
    function GetOnHide: TNotifyEvent;
    function GetOnShow: TNotifyEvent;
    function GetPosition: TPosition;
    function GetWindowState: TWindowState;
    procedure SetBorderIcons(AValue: TBorderIcons);
    procedure SetBorderStyle(AValue: TFormBorderStyle);
    procedure SetDesignTimePPI(AValue: Integer);
    procedure SetFormStyle(AValue: TFormStyle);
    procedure SetKeyPreview(AValue: Boolean);
    procedure SetLCLVersion(AValue: string);
    procedure SetModalResult(AValue: TModalResult);
    procedure SetOnCloseQuery(AValue: TCloseQueryEvent);
    procedure SetOnCreate(AValue: TNotifyEvent);
    procedure SetOnDeactivate(AValue: TNotifyEvent);
    procedure SetOnDestroy(AValue: TNotifyEvent);
    procedure SetOnHide(AValue: TNotifyEvent);
    procedure SetOnShow(AValue: TNotifyEvent);
    procedure DoOnCreate;
    procedure FormMessageHandler(var TheMessage: TLMessage);
    procedure SetPosition(AValue: TPosition);
    procedure SetWindowState(AValue: TWindowState);
    procedure ActivateHandler(Sender: TObject);
    procedure DeactivateHandler(Sender: TObject);
    procedure DoClose(var CloseAction: TCloseAction); virtual;
    procedure CloseHandler(Sender: TObject; var CloseAction: TCloseAction);
    procedure CloseQueryHandler(Sender : TObject; var CanClose : boolean);
  protected
    procedure CreateParams(var p: TCreateParams); virtual;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function GetNativeScrollingWinControl: TScrollingWinControl; override;
    function GetNativeForm: TForm; virtual;
    procedure UpdateNativeControl; override;
  public
    NativeForm: TForm;
    procedure AfterConstruction; override;
    property ModalResult: TModalResult read GetModalResult write SetModalResult;
    function ShowModal: Integer; virtual;
    procedure Close;
    function CloseQuery: boolean; virtual;
    procedure BringToFront;
    procedure Release;
    constructor Create(TheOwner: TComponent); override;
    constructor CreateNew(AOwner: TComponent; Num: Integer = 0); virtual;
    destructor Destroy; override;
  published
    property DesignTimePPI: Integer read GetDesignTimePPI write SetDesignTimePPI; // Not used
    property FormState: TFormState read GetFormState;
    property FormStyle: TFormStyle read GetFormStyle write SetFormStyle;
    property BorderStyle: TFormBorderStyle read GetBorderStyle write SetBorderStyle default bsSizeable;
    property BorderIcons: TBorderIcons read GetBorderIcons write SetBorderIcons;
    property LCLVersion: string read GetLCLVersion write SetLCLVersion;
    property KeyPreview: Boolean read GetKeyPreview write SetKeyPreview default False;
    property Position: TPosition read GetPosition write SetPosition default poDesigned;
    property WindowState: TWindowState read GetWindowState write SetWindowState default wsNormal;
    property OnShow: TNotifyEvent read GetOnShow write SetOnShow;
    property OnHide: TNotifyEvent read GetOnHide write SetOnHide;
    property OnCreate: TNotifyEvent read GetOnCreate write SetOnCreate;
    property OnDestroy: TNotifyEvent read GetOnDestroy write SetOnDestroy;
    property OnActivate: TNotifyEvent read FOnActivate write FOnActivate;
    property OnDeactivate: TNotifyEvent read FOnDeactivate write FOnDeactivate;
    property OnClose: TCloseEvent read FOnClose write FOnClose;
    property OnCloseQuery: TCloseQueryEvent read FOnCloseQuery write FOnCloseQuery;
    property ClientHeight;
    property ClientWidth;
    property OnMouseUp;
    property OnMouseDown;
    property OnMouseMove;
  end;

  TDpiForms = specialize TFPGObjectList<TDpiForm>;

  { TDpiEdit }

  TDpiEdit = class(TDpiWinControl)
  private
    FText: string;
    function GetSelLength: Integer;
    function GetSelStart: Integer;
    procedure SetSelLength(AValue: Integer);
    procedure SetSelStart(AValue: Integer);
  protected
    function GetNativeWinControl: TWinControl; override;
    function GetNativeEdit: TEdit; virtual;
    function GetText: TCaption; override;
    procedure SetText(AValue: TCaption); override;
  public
    NativeEdit: TEdit;
    destructor Destroy; override;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelStart: Integer read GetSelStart write SetSelStart;
  published
    property Text;
    property BorderStyle default bsSingle;
    property ParentFont;
  end;

  { TDpiButton }

  TDpiButton = class(TDpiControl)
  private
  protected
    function GetNativeControl: TControl; override;
  public
    NativeButton: TButton;
    destructor Destroy; override;
  published
    property Visible;
  end;

  { TDpiListBox }

  TDpiListBox = class(TDpiWinControl)
  private
    function GetBorderStyle: TBorderStyle;
    function GetCount: Integer;
    function GetExtendedSelect: Boolean;
    function GetIntegralHeight: Boolean;
    function GetItemHeight: Integer;
    function GetItemIndex: Integer;
    function GetItems: TStrings;
    function GetParentFont: Boolean;
    function GetScrollWidth: Integer;
    function GetTopIndex: Integer;
    procedure SetBorderStyle(AValue: TBorderStyle);
    procedure SetExtendedSelect(AValue: Boolean);
    procedure SetIntegralHeight(AValue: Boolean);
    procedure SetItemHeight(AValue: Integer);
    procedure SetItemIndex(AValue: Integer);
    procedure SetItems(AValue: TStrings);
    procedure SetParentFont(AValue: Boolean);
    procedure SetScrollWidth(AValue: Integer);
    procedure SetTopIndex(AValue: Integer);
  protected
    function GetNativeWinControl: TWinControl; override;
    function GetNativeListBox: TListBox; virtual;
  public
    NativeListBox: TListBox;
    destructor Destroy; override;
    property ItemIndex: Integer read GetItemIndex write SetItemIndex;
    property Items: TStrings read GetItems write SetItems;
    property Count: Integer read GetCount;
  published
    property TopIndex: Integer read GetTopIndex write SetTopIndex default 0;
    property ScrollWidth: Integer read GetScrollWidth write SetScrollWidth default 0;
    property ParentFont: Boolean read GetParentFont write SetParentFont default True;
    property ItemHeight: Integer read GetItemHeight write SetItemHeight;
    property IntegralHeight: Boolean read GetIntegralHeight write SetIntegralHeight default False;
    property ExtendedSelect: Boolean read GetExtendedSelect write SetExtendedSelect default true;
    property BorderStyle: TBorderStyle read GetBorderStyle write SetBorderStyle default bsNone;
    property Visible;
    property Anchors;
  end;

  { TDpiScrollBar }

  TDpiScrollBar = class(TDpiControl)
  private
    function GetKind: TScrollBarKind;
    function GetMax: Integer;
    function GetMin: Integer;
    function GetOnChange: TNotifyEvent;
    function GetPageSize: Integer;
    function GetPosition: Integer;
    procedure SetKind(AValue: TScrollBarKind);
    procedure SetMax(AValue: Integer);
    procedure SetMin(AValue: Integer);
    procedure SetOnChange(AValue: TNotifyEvent);
    procedure SetPageSize(AValue: Integer);
    procedure SetPosition(AValue: Integer);
  protected
    function GetNativeControl: TControl; override;
  public
    NativeScrollBar: TScrollBar;
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
  published
    property PageSize: Integer read GetPageSize write SetPageSize;
    property Min: Integer read GetMin write SetMin;
    property Max: Integer read GetMax write SetMax;
    property Position: Integer read GetPosition write SetPosition;
    property Kind: TScrollBarKind read GetKind write SetKind;
    property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
    property Visible;
  end;

  { TDpiBitmap }

  TDpiBitmap = class(TDpiRasterImage)
  private
    FCanvas: TDpiCanvas;
    FWidth: Integer;
    FHeight: Integer;
    function GetCanvas: TDpiCanvas;
    function GetPixelFormat: TPixelFormat;
    function GetScanLine(Row: Integer): Pointer;
    procedure SetPixelFormat(AValue: TPixelFormat);
  protected
    function GetHeight: Integer; override;
    function GetWidth: Integer; override;
    function GetNativeBitmap: TCustomBitmap; virtual;
    function GetNativeRasterImage: TRasterImage; override;
    procedure SetHeight(AValue: Integer); override;
    procedure SetWidth(AValue: Integer); override;
    procedure ScreenChanged; override;
  public
    NativeBitmap: TBitmap;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure SetSize(AWidth, AHeight: Integer);
    constructor Create; override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property ScanLine[Row: Integer]: Pointer read GetScanLine;
  published
    property PixelFormat: TPixelFormat read GetPixelFormat write SetPixelFormat;
    property Height: Integer read GetHeight write SetHeight;
    property Width: Integer read GetWidth write SetWidth;
    property Canvas: TDpiCanvas read GetCanvas;
  end;

  { TDpiPicture }

  TDpiPicture = class(TPersistent)
  private
    FBitmap: TDpiBitmap;
    procedure SetBitmap(AValue: TDpiBitmap);
  published
    procedure LoadFromFile(FileName: string);
    property Bitmpa: TDpiBitmap read FBitmap write SetBitmap;
  end;

  { TDpiImage }

  TDpiImage = class(TDpiControl)
  private
    FDpiPicture: TDpiPicture;
    FStretch: Boolean;
    procedure SetPicture(AValue: TDpiPicture);
    procedure SetStretch(AValue: Boolean);
  protected
  public
    NativeImage: TImage;
    function GetNativeControl: TControl; override;
    destructor Destroy; override;
  published
    property Stretch: Boolean read FStretch write SetStretch;
    property Picture: TDpiPicture read FDpiPicture write SetPicture;
    property Visible;
  end;

  { TDpiPaintBox }

  TDpiPaintBox = class(TDpiGraphicControl)
  public
    NativePaintBox: TPaintBox;
    function GetNativeGraphicControl: TGraphicControl; override;
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
  published
    property OnPaint;
    property Visible;
  end;

  { TDpiScreen }

  TDpiScreen = class
  private
    // TScreen
    FDpi: Integer;
    FActiveFormPrev: TDpiForm;
    FActiveForm: TDpiForm;
    FPrevActiveForms: TDpiForms;
    FForms: TDpiForms;
    procedure AddForm(AForm: TDpiForm);
    procedure RemoveForm(AForm: TDpiForm);
    function GetActiveForm: TDpiForm;
    function GetCursor: TCursor;
    function GetCursors(Index: Integer): HCURSOR;
    function GetFormCount: Integer;
    function GetForms(Index: Integer): TDpiForm;
    function GetHeight: Integer;
    function GetWidth: Integer;
    procedure SetCursor(AValue: TCursor);
    procedure SetCursors(Index: Integer; AValue: HCURSOR);
    procedure SetDpi(AValue: Integer);
    procedure UpdateForms;
  public
    constructor Create;
    destructor Destroy; override;
    procedure UpdateScreen;
    procedure UpdateActiveFormFromNativeScreen;
    property FormCount: Integer read GetFormCount;
    property Forms[Index: Integer]: TDpiForm read GetForms;
    property ActiveForm: TDpiForm read GetActiveForm;
    property Cursor: TCursor read GetCursor write SetCursor;
    property Cursors[Index: Integer]: HCURSOR read GetCursors write SetCursors;
  published
    property Dpi: Integer read FDpi write SetDpi;
    property PixelsPerInch: Integer read FDpi;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
  end;

  { TDpiJpegImage }

  TDpiJpegImage = class(TDpiBitmap)
  protected
    function GetNativeBitmap: TCustomBitmap; override;
    function GetNativeJpeg: TJPEGImage; virtual;
  public
    NativeJpeg: TJPEGImage;
    constructor Create; override;
    destructor Destroy; override;
  end;

  { TDpiPortableNetworkGraphic }

  TDpiPortableNetworkGraphic = class(TDpiBitmap)
  protected
    function GetNativeBitmap: TCustomBitmap; override;
    function GetNativePng: TPortableNetworkGraphic; virtual;
  public
    NativePng: TPortableNetworkGraphic;
    constructor Create; override;
    destructor Destroy; override;
  end;

  { TDpiApplication }

  TDpiApplication = class(TComponent)
  private
    FMainForm: TDpiForm;
    FCreatingForm: TDpiForm;
    FOldExitProc: Pointer;
    function GetActive: Boolean;
    function GetShowMainForm: Boolean;
    function GetTitle: string;
    procedure SetMainForm(AValue: TDpiForm);
    function GetMainForm: TDpiForm;
    procedure SetShowMainForm(AValue: Boolean);
    procedure SetTitle(AValue: string);
  protected
    function GetNativeApplication: TApplication; virtual;
    procedure DoBeforeFinalization;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Run;
    procedure Initialize;
    procedure Terminate;
    procedure ProcessMessages;
    procedure UpdateMainForm(AForm: TDpiForm);
    procedure CreateForm(InstanceClass: TComponentClass; out Reference);
    function MessageBox(Text, Caption: PChar; Flags: Longint = MB_OK): Integer;
    property MainForm: TDpiForm read GetMainForm write SetMainForm;
    property ShowMainForm: Boolean read GetShowMainForm write SetShowMainForm default True;
    property Title: string read GetTitle write SetTitle;
    property Active: Boolean read GetActive;
  end;

  { TDpiMouse }

  // TMouse
  TDpiMouse = class
  private
    function GetCursorPos: TPoint;
    procedure SetCursorPos(AValue: TPoint);
  public
    constructor Create;
    property CursorPos: TPoint read GetCursorPos write SetCursorPos;
  end;

  { TDpiMenuItem }

  TDpiMenuItem = class(TComponent)
  private
    FItems: TList;
    FParent: TDpiMenuItem;
    FOnClick: TNotifyEvent;
    function GetCaption: TTranslateString;
    function GetChecked: Boolean;
    function GetCount: Integer;
    function GetEnabled: Boolean;
    function GetGroupIndex: Byte;
    function GetItem(Index: Integer): TDpiMenuItem;
    function GetOnClick: TNotifyEvent;
    function GetRadioItem: Boolean;
    function GetShortCut: TShortCut;
    function GetVisible: Boolean;
    function IsCaptionStored: Boolean;
    function IsCheckedStored: Boolean;
    function IsEnabledStored: Boolean;
    function IsShortCutStored: Boolean;
    function IsVisibleStored: Boolean;
    procedure SetCaption(AValue: TTranslateString);
    procedure SetChecked(AValue: Boolean);
    procedure SetEnabled(AValue: Boolean);
    procedure SetGroupIndex(AValue: Byte);
    procedure SetOnClick(AValue: TNotifyEvent);
    procedure SetRadioItem(AValue: Boolean);
    procedure SetShortCut(AValue: TShortCut);
    procedure SetVisible(AValue: Boolean);
    procedure OnClickHandler(Sender: TObject);
  protected
    function GetNativeMenuItem: TMenuItem; virtual;
    procedure SetParentComponent(AValue: TComponent); override;
  public
    NativeMenuItem: TMenuItem;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Delete(Index: Integer);
    procedure Add(Item: TDpiMenuItem);
    procedure Insert(Index: Integer; Item: TDpiMenuItem);
    function IndexOf(Item: TDpiMenuItem): Integer;
    procedure Remove(Item: TDpiMenuItem);
    property Items[Index: Integer]: TDpiMenuItem read GetItem; default;
    property Count: Integer read GetCount;
  published
    property RadioItem: Boolean read GetRadioItem write SetRadioItem default False;
    property ShortCut: TShortCut read GetShortCut write SetShortCut
      stored IsShortCutStored default 0;
    property Enabled: Boolean read GetEnabled write SetEnabled
      stored IsEnabledStored default True;
    property Visible: Boolean read GetVisible write SetVisible
      stored IsVisibleStored default True;
    property Checked: Boolean read GetChecked write SetChecked
      stored IsCheckedStored default False;
    property Caption: TTranslateString read GetCaption write SetCaption
      stored IsCaptionStored;
    property OnClick: TNotifyEvent read GetOnClick write SetOnClick;
    property GroupIndex: Byte read GetGroupIndex write SetGroupIndex default 0;
  end;

  TDpiMenu = class(TComponent)
  private
    FItems: TDpiMenuItem;
  protected
    function GetNativeMenu: TMenu; virtual;
  public
    property Items: TDpiMenuItem read FItems;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  { TDpiPopupMenu }

  TDpiPopupMenu = class(TDpiMenu)
  private
    function GetAutoPopup: Boolean;
    procedure SetAutoPopup(AValue: Boolean);
  protected
    function GetNativeMenu: TMenu; override;
    function GetNativePopupMenu: TPopupMenu; virtual;
  public
    NativePopupMenu: TPopupMenu;
    procedure PopUp;
    procedure PopUp(X, Y: Integer); virtual;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property AutoPopup: Boolean read GetAutoPopup write SetAutoPopup default True;
  end;

var
  DpiFormFileDesc: TDpiFormFileDesc;
  DpiScreen: TDpiScreen;
  DpiApplication: TDpiApplication;
  DpiMouse: TDpiMouse;

procedure Register;
function DpiBitBlt(DestDC: HDC; X, Y, Width, Height: Integer; SrcDC: HDC; XSrc, YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
function DpiBitBltCanvas(Dest: TDpiCanvas; X, Y, Width, Height: Integer; Src: TDpiCanvas; XSrc, YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
function DpiCreateRectRgn(X1, Y1, X2, Y2: Integer): HRGN;
function DpiScrollDC(Canvas: TDpiCanvas; dx: Longint; dy: Longint; const lprcScroll: TRect;
  const lprcClip:TRect; hrgnUpdate: Handle; lprcUpdate: PRect): Boolean;
function ScaleToNative(Value: Integer): Integer;
function ScaleToNativeDist(Base, Value: Integer): Integer;
function ScaleFromNative(Value: Integer): Integer;
function ScalePointToNative(Value: TPoint): TPoint;
function ScalePointFromNative(Value: TPoint): TPoint;
function ScaleSizeToNative(Value: TSize): TSize;
function ScaleSizeFromNative(Value: TSize): TSize;
function ScaleRectToNative(Value: TRect): TRect;
function ScaleRectFromNative(Value: TRect): TRect;
function ScaleFloatToNative(Value: Double): Double;
function ScaleFloatFromNative(Value: Double): Double;
procedure WriteLog(Text: string);
function DpiGetSystemMetrics(nIndex: Integer): Integer;

implementation

uses
  LCLStrConsts;

resourcestring
  SDpiFormTitle = 'DpiForm form';
  SDpiFormDescription = 'DPI aware form';

procedure Register;
begin
  FormEditingHook.RegisterDesignerBaseClass(TDpiForm);
  DpiFormFileDesc := TDpiFormFileDesc.Create;
  RegisterProjectFileDescriptor(DpiFormFileDesc);
  RegisterComponents('DpiControls', [TDpiButton, TDpiImage, TDpiPaintBox,
    TDpiListBox, TDpiPopupMenu, TDpiEdit]);
end;

function DpiBitBltCanvas(Dest: TDpiCanvas; X, Y, Width, Height: Integer;
  Src: TDpiCanvas; XSrc, YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
begin
  Result := DpiBitBlt(Dest.Handle, X, Y, Width, Height, Src.Handle, XSrc, YSrc, Rop);
end;

function DpiCreateRectRgn(X1, Y1, X2, Y2: Integer): HRGN;
begin
  Result := CreateRectRgn(ScaleToNative(X1), ScaleToNative(Y1), ScaleToNative(X2),
    ScaleToNative(Y2));
end;

function DpiScrollDC(Canvas: TDpiCanvas; dx: Longint; dy: Longint;
  const lprcScroll: TRect; const lprcClip: TRect; hrgnUpdate: Handle;
  lprcUpdate: PRect): Boolean;
begin
  {$IFDEF WINDOWS}
  Result := Windows.ScrollDC(Canvas.Handle, ScaleToNative(dx), ScaleToNative(dy),
    ScaleRectToNative(lprcScroll), ScaleRectToNative(lprcClip),
    hrgnUpdate, lprcUpdate);
  {$ENDIF}
  {$IFDEF LINUX}
  // Can't do scrolling of DC under Linux, then fallback into BitBlt.
  Result := DpiBitBltCanvas(Canvas, lprcScroll.Left + dx, lprcScroll.Top + dy, lprcScroll.Right - lprcScroll.Left, lprcScroll.Bottom - lprcScroll.Top,
    Canvas, lprcScroll.Left, lprcScroll.Top);
  {$ENDIF}
end;

function Ceil(const X: Single): Integer;
begin
  if X > High(Integer) then
    Result := High(Integer)
  else if X < Low(Integer) then
    Result := Low(Integer)
  else begin
    Result := Trunc(X);
    if (Result <> X) then begin
      if (Result > 0) then Inc(Result) else Dec(Result);
    end;
  end;
end;

function ScaleToNative(Value: Integer): Integer;
begin
  Result := Round(Value * DpiScreen.Dpi / 96);
end;

function ScaleToNativeDist(Base, Value: Integer): Integer;
begin
  Result := ScaleToNative(Base + Value) - ScaleToNative(Base);
end;

function ScaleFromNative(Value: Integer): Integer;
begin
  Result := Round(Value * 96 / DpiScreen.Dpi);
end;

function ScalePointToNative(Value: TPoint): TPoint;
begin
  Result.X := ScaleToNative(Value.X);
  Result.Y := ScaleToNative(Value.Y);
end;

function ScalePointFromNative(Value: TPoint): TPoint;
begin
  Result.X := ScaleFromNative(Value.X);
  Result.Y := ScaleFromNative(Value.Y);
end;

function ScaleSizeToNative(Value: TSize): TSize;
begin
  Result.Width := ScaleToNative(Value.Width);
  Result.Height := ScaleToNative(Value.Height);
end;

function ScaleSizeFromNative(Value: TSize): TSize;
begin
  Result.Width := ScaleFromNative(Value.Width);
  Result.Height := ScaleFromNative(Value.Height);
end;

function ScaleRectToNative(Value: TRect): TRect;
begin
  Result.Left := ScaleToNative(Value.Left);
  Result.Top := ScaleToNative(Value.Top);
  Result.Right := ScaleToNative(Value.Right);
  Result.Bottom := ScaleToNative(Value.Bottom);
end;

function ScaleRectFromNative(Value: TRect): TRect;
begin
  Result.Left := ScaleFromNative(Value.Left);
  Result.Top := ScaleFromNative(Value.Top);
  Result.Right := ScaleFromNative(Value.Right);
  Result.Bottom := ScaleFromNative(Value.Bottom);
end;

function ScaleFloatToNative(Value: Double): Double;
begin
  Result := Value * DpiScreen.Dpi / 96;
end;

function ScaleFloatFromNative(Value: Double): Double;
begin
  Result := Value * 96 / DpiScreen.Dpi;
end;

procedure WriteLog(Text: string);
var
  F: Text;
const
  FileName = 'Log.txt';
begin
  AssignFile(F, FileName);
  if FileExists(FileName) then Append(F) else Rewrite(F);
  WriteLn(F, Text);
  CloseFile(F);
end;

function DpiGetSystemMetrics(nIndex: Integer): Integer;
begin
  Result := ScaleFromNative(GetSystemMetrics(nIndex));
end;

function DpiBitBlt(DestDC: HDC; X, Y, Width, Height: Integer; SrcDC: HDC; XSrc,
  YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
var
  DstWidth, DstHeight: Integer;
  SrcWidth, SrcHeight: Integer;
begin
  {$IFDEF WINDOWS}
  // LCLIntf.BitBlt is slower than direct Windows BitBlt
  Result := Windows.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
    ScaleToNative(Width), ScaleToNative(Height), SrcDC,
    ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);
  {$ELSE}


  DstWidth := ScaleToNativeDist(X, Width);
  DstHeight := ScaleToNativeDist(Y, Height);
  SrcWidth := ScaleToNativeDist(XSrc, Width);
  SrcHeight := ScaleToNativeDist(YSrc, Height);
  if (DstWidth = SrcWidth) and (DstHeight = SrcHeight) then begin
    Result := BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      DstWidth, DstHeight, SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);
  end else begin
    Result := StretchBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      DstWidth, DstHeight, SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc),
      SrcWidth, SrcHeight, Rop);
  end;

{  Result := BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
    ScaleToNative(Width), ScaleToNative(Height), SrcDC,
    ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);
 }
  {$ENDIF}
end;

{ TDpiControlBorderSpacing }

function TDpiControlBorderSpacing.IsBottomStored: Boolean;
begin
  if FDefault = nil
  then Result := FBottom <> 0
  else Result := FBottom <> FDefault^.Bottom;
end;

function TDpiControlBorderSpacing.IsAroundStored: Boolean;
begin
  if FDefault = nil
  then Result := FAround <> 0
  else Result := FAround <> FDefault^.Around;
end;

function TDpiControlBorderSpacing.IsLeftStored: Boolean;
begin
  if FDefault = nil
  then Result := FLeft <> 0
  else Result := FLeft <> FDefault^.Left;
end;

function TDpiControlBorderSpacing.IsRightStored: Boolean;
begin
  if FDefault = nil
  then Result := FRight <> 0
  else Result := FRight <> FDefault^.Right;
end;

function TDpiControlBorderSpacing.IsTopStored: Boolean;
begin
  if FDefault = nil
  then Result := FTop <> 0
  else Result := FTop <> FDefault^.Top;
end;

procedure TDpiControlBorderSpacing.SetAround(AValue: TSpacingSize);
begin
  if FAround = AValue then Exit;
  FAround := AValue;
  Change(False);
end;

procedure TDpiControlBorderSpacing.SetBottom(AValue: TSpacingSize);
begin
  if FBottom = AValue then Exit;
  FBottom := AValue;
  Change(False);
end;

procedure TDpiControlBorderSpacing.SetLeft(AValue: TSpacingSize);
begin
  if FLeft = AValue then Exit;
  FLeft := AValue;
  Change(False);
end;

procedure TDpiControlBorderSpacing.SetRight(AValue: TSpacingSize);
begin
  if FRight = AValue then Exit;
  FRight := AValue;
  Change(False);
end;

procedure TDpiControlBorderSpacing.SetTop(AValue: TSpacingSize);
begin
  if FTop = AValue then Exit;
  FTop := AValue;
  Change(False);
end;

procedure TDpiControlBorderSpacing.Change(InnerSpaceChanged: Boolean);
begin
  if FControl <> nil then
    FControl.DoBorderSpacingChange(Self, InnerSpaceChanged);
  if Assigned(OnChange) then OnChange(Self);
end;

constructor TDpiControlBorderSpacing.Create(OwnerControl: TDpiControl;
  ADefault: PControlBorderSpacingDefault);
begin
  FControl := OwnerControl;
  FDefault := ADefault;
  if ADefault <> nil then
  begin
    FLeft := ADefault^.Left;
    FRight := ADefault^.Right;
    FTop := ADefault^.Top;
    FBottom := ADefault^.Bottom;
    FAround := ADefault^.Around;
  end;
  inherited Create;
end;

{ TDpiEdit }

function TDpiEdit.GetText: TCaption;
begin
  Result := FText;
end;

procedure TDpiEdit.SetText(AValue: TCaption);
begin
  FText := AValue;
  GetNativeEdit.Text := AValue;
end;

destructor TDpiEdit.Destroy;
begin
  FreeAndNil(NativeEdit);
  inherited;
end;

function TDpiEdit.GetSelLength: Integer;
begin
  Result := GetNativeEdit.SelLength;
end;

function TDpiEdit.GetSelStart: Integer;
begin
  Result := GetNativeEdit.SelStart;
end;

procedure TDpiEdit.SetSelLength(AValue: Integer);
begin
  GetNativeEdit.SelLength := AValue;
end;

procedure TDpiEdit.SetSelStart(AValue: Integer);
begin
  GetNativeEdit.SelStart := AValue;
end;

function TDpiEdit.GetNativeWinControl: TWinControl;
begin
  Result := GetNativeEdit;
end;

function TDpiEdit.GetNativeEdit: TEdit;
begin
  if not Assigned(NativeEdit) then NativeEdit := TEdit.Create(nil);
  Result := NativeEdit;
end;

{ TDpiMenu }

function TDpiMenu.GetNativeMenu: TMenu;
begin
  Result := nil;
end;

constructor TDpiMenu.Create(AOwner: TComponent);
begin
  inherited;
  FItems := TDpiMenuItem.Create(Self);
end;

destructor TDpiMenu.Destroy;
begin
  FreeAndNil(FItems);
  inherited;
end;

{ TDpiMenuItem }

function TDpiMenuItem.GetCaption: TTranslateString;
begin
  Result := GetNativeMenuItem.Caption;
end;

function TDpiMenuItem.GetChecked: Boolean;
begin
  Result := GetNativeMenuItem.Checked;
end;

function TDpiMenuItem.GetCount: Integer;
begin
  Result := FItems.Count;
end;

function TDpiMenuItem.GetEnabled: Boolean;
begin
  Result := GetNativeMenuItem.Enabled;
end;

function TDpiMenuItem.GetGroupIndex: Byte;
begin
  Result := GetNativeMenuItem.GroupIndex;
end;

function TDpiMenuItem.GetItem(Index: Integer): TDpiMenuItem;
begin
  Result := TDpiMenuItem(FItems[Index]);
end;

function TDpiMenuItem.GetOnClick: TNotifyEvent;
begin
  Result := FOnClick;
end;

function TDpiMenuItem.GetRadioItem: Boolean;
begin
  Result := GetNativeMenuItem.RadioItem;
end;

function TDpiMenuItem.GetShortCut: TShortCut;
begin
  Result := GetNativeMenuItem.ShortCut;
end;

function TDpiMenuItem.GetVisible: Boolean;
begin
  Result := GetNativeMenuItem.Visible;
end;

function TDpiMenuItem.IsCaptionStored: Boolean;
begin
  Result := False;
end;

function TDpiMenuItem.IsCheckedStored: Boolean;
begin
  Result := False;
end;

function TDpiMenuItem.IsEnabledStored: Boolean;
begin
  Result := False;
end;

function TDpiMenuItem.IsShortCutStored: Boolean;
begin
  Result := False;
end;

function TDpiMenuItem.IsVisibleStored: Boolean;
begin
  Result := False;
end;

procedure TDpiMenuItem.SetCaption(AValue: TTranslateString);
begin
  GetNativeMenuItem.Caption := AValue;
end;

procedure TDpiMenuItem.SetChecked(AValue: Boolean);
begin
  GetNativeMenuItem.Checked := AValue;
end;

procedure TDpiMenuItem.SetEnabled(AValue: Boolean);
begin
  GetNativeMenuItem.Enabled := AValue;
end;

procedure TDpiMenuItem.SetGroupIndex(AValue: Byte);
begin
  GetNativeMenuItem.GroupIndex := AValue;
end;

procedure TDpiMenuItem.SetOnClick(AValue: TNotifyEvent);
begin
  FOnClick := AValue;
end;

procedure TDpiMenuItem.SetRadioItem(AValue: Boolean);
begin
  GetNativeMenuItem.RadioItem := AValue;
end;

procedure TDpiMenuItem.SetShortCut(AValue: TShortCut);
begin
  GetNativeMenuItem.ShortCut := AValue;
end;

procedure TDpiMenuItem.SetVisible(AValue: Boolean);
begin
  GetNativeMenuItem.Visible := AValue;
end;

procedure TDpiMenuItem.OnClickHandler(Sender: TObject);
begin
  if Assigned(FOnClick) then
    FOnClick(Self);
end;

procedure TDpiMenuItem.Delete(Index: Integer);
begin
  FItems.Delete(Index);
  GetNativeMenuItem.Delete(Index);
end;

procedure TDpiMenuItem.Add(Item: TDpiMenuItem);
begin
  Insert(GetCount, Item);
end;

procedure TDpiMenuItem.Insert(Index: Integer; Item: TDpiMenuItem);
begin
  FItems.Insert(Index, Item);
  GetNativeMenuItem.Insert(Index, Item.GetNativeMenuItem);
end;

function TDpiMenuItem.IndexOf(Item: TDpiMenuItem): Integer;
begin
  if FItems = nil then
    Result := -1
  else
    Result := FItems.IndexOf(Item);
end;

procedure TDpiMenuItem.Remove(Item: TDpiMenuItem);
var
  I: Integer;
begin
  I := IndexOf(Item);
  if I < 0 then
    raise EMenuError.Create(SMenuNotFound);
  Delete(I);
end;

function TDpiMenuItem.GetNativeMenuItem: TMenuItem;
begin
  if not Assigned(NativeMenuItem) then begin
    NativeMenuItem := TMenuItem.Create(nil);
    NativeMenuItem.Name := 'Native' + Name;
    NativeMenuItem.OnClick := @OnClickHandler;
  end;
  Result := NativeMenuItem;
end;

procedure TDpiMenuItem.SetParentComponent(AValue: TComponent);
begin
  if (FParent = AValue) then Exit;
  if Assigned(FParent) then FParent.Remove(Self);
  if Assigned(AValue) then
  begin
    if (AValue is TDpiMenu)
      then TDpiMenu(AValue).Items.Add(Self)
    else if (AValue is TDpiMenuItem)
      then TDpiMenuItem(AValue).Add(Self)
    else
      raise Exception.Create('TDpiMenuItem.SetParentComponent: suggested parent not of type TDpiMenu or TDpiMenuItem');
   end;
end;

constructor TDpiMenuItem.Create(AOwner: TComponent);
begin
  inherited;
  FItems := TList.Create;
end;

destructor TDpiMenuItem.Destroy;
begin
  FreeAndNil(FItems);
  // TODO: Release menu items
  //FreeAndNil(NativeMenuItem);
  inherited;
end;

{ TDpiPopupMenu }

procedure TDpiPopupMenu.PopUp;
var
  Pos: TPoint;
begin
  Pos := DpiMouse.CursorPos;
  Popup(Pos.X, Pos.Y);
end;

procedure TDpiPopupMenu.PopUp(X, Y: Integer);
begin
  GetNativePopupMenu.PopUp(ScaleToNative(X), ScaleToNative(Y));
end;

constructor TDpiPopupMenu.Create(AOwner: TComponent);
begin
  inherited;
  GetNativePopupMenu;
end;

function TDpiPopupMenu.GetAutoPopup: Boolean;
begin
  Result := GetNativePopupMenu.AutoPopup;
end;

procedure TDpiPopupMenu.SetAutoPopup(AValue: Boolean);
begin
  GetNativePopupMenu.AutoPopup := AValue;
end;

function TDpiPopupMenu.GetNativeMenu: TMenu;
begin
  Result := GetNativePopupMenu;
end;

function TDpiPopupMenu.GetNativePopupMenu: TPopupMenu;
begin
  if not Assigned(NativePopupMenu) then begin
    NativePopupMenu := TPopupMenu.Create(nil);
    if Assigned(Items.NativeMenuItem) then Items.NativeMenuItem.Free;
    Items.NativeMenuItem := NativePopupMenu.Items;
  end;
  Result := NativePopupMenu;
end;

destructor TDpiPopupMenu.Destroy;
begin
  if Assigned(NativePopupMenu) then FreeAndNil(NativePopupMenu);
  inherited;
end;

{ TDpiMouse }

function TDpiMouse.GetCursorPos: TPoint;
begin
  Result := ScalePointFromNative(Mouse.CursorPos);
end;

procedure TDpiMouse.SetCursorPos(AValue: TPoint);
begin
  Mouse.CursorPos := ScalePointToNative(AValue);
end;

constructor TDpiMouse.Create;
begin

end;

{ TDpiSizeConstraints }

procedure TDpiSizeConstraints.SetMaxHeight(AValue: TConstraintSize);
begin
  if FMaxHeight=AValue then Exit;
  FMaxHeight:=AValue;
end;

procedure TDpiSizeConstraints.SetMaxWidth(AValue: TConstraintSize);
begin
  if FMaxWidth=AValue then Exit;
  FMaxWidth:=AValue;
end;

procedure TDpiSizeConstraints.SetMinHeight(AValue: TConstraintSize);
begin
  if FMinHeight=AValue then Exit;
  FMinHeight:=AValue;
end;

procedure TDpiSizeConstraints.SetMinWidth(AValue: TConstraintSize);
begin
  if FMinWidth=AValue then Exit;
  FMinWidth:=AValue;
end;

{ TDpiScrollingWinControl }

function TDpiScrollingWinControl.GetNativeCustomControl: TCustomControl;
begin
  Result := GetNativeScrollingWinControl;
end;

function TDpiScrollingWinControl.GetNativeScrollingWinControl: TScrollingWinControl;
begin
  Result := nil;
end;

constructor TDpiScrollingWinControl.Create(TheOwner: TComponent);
begin
  inherited;
  FHorzScrollBar := TDpiControlScrollBar.Create;
  FVertScrollBar := TDpiControlScrollBar.Create;
end;

destructor TDpiScrollingWinControl.Destroy;
begin
  FreeAndNil(FHorzScrollBar);
  FreeAndNil(FVertScrollBar);
  inherited;
end;

{ TDpiControlScrollBar }

function TDpiControlScrollBar.GetVisible: Boolean;
begin

end;

procedure TDpiControlScrollBar.SetVisible(AValue: Boolean);
begin

end;

{ TGraphicControlEx }

procedure TGraphicControlEx.Paint;
begin
  inherited Paint;
end;

{ TFormEx }

procedure TFormEx.WndProc(var TheMessage: TLMessage);
begin
  inherited WndProc(TheMessage);
  if Assigned(FOnMessage) then
    FOnMessage(TheMessage);
end;

{ TDpiApplication }

procedure TDpiApplication.SetMainForm(AValue: TDpiForm);
begin
  FMainForm := AValue;
end;

function TDpiApplication.GetTitle: string;
begin
  Result := GetNativeApplication.Title;
end;

function TDpiApplication.GetShowMainForm: Boolean;
begin
  Result := GetNativeApplication.ShowMainForm;
end;

function TDpiApplication.GetActive: Boolean;
begin
  Result := Application.Active;
end;

function TDpiApplication.GetMainForm: TDpiForm;
begin
  Result := FMainForm;
end;

procedure TDpiApplication.SetShowMainForm(AValue: Boolean);
begin
  GetNativeApplication.ShowMainForm := AValue;
end;

procedure TDpiApplication.SetTitle(AValue: string);
begin
  GetNativeApplication.Title := AValue;
end;

function TDpiApplication.GetNativeApplication: TApplication;
begin
  Result := Application;
end;

procedure DpiBeforeFinalization;
// This is our ExitProc handler.
begin
  DpiApplication.DoBeforeFinalization;
end;

procedure TDpiApplication.DoBeforeFinalization;
var
  I: Integer;
begin
  if Self = nil then Exit;
  for I := ComponentCount - 1 downto 0 do begin
    if I < ComponentCount then
      Components[I].Free;
  end;
end;

function DpiFindApplicationComponent(const ComponentName: string): TComponent;
// Note: this function is used by TReader to auto rename forms to unique names.
begin
  Result := DpiApplication.FindComponent(ComponentName);
end;

constructor TDpiApplication.Create(AOwner: TComponent);
begin
  RegisterFindGlobalComponentProc(@DpiFindApplicationComponent);
  FOldExitProc := ExitProc;
  ExitProc := @DpiBeforeFinalization;
  inherited;
end;

destructor TDpiApplication.Destroy;
begin
  UnregisterFindGlobalComponentProc(@DpiFindApplicationComponent);
  ExitProc := FOldExitProc;
  inherited;
end;

procedure TDpiApplication.Run;
begin
  if (FMainForm <> nil) and GetShowMainForm then FMainForm.Show;
  GetNativeApplication.Run;
end;

procedure TDpiApplication.Initialize;
begin
  GetNativeApplication.Initialize;
  DpiScreen.UpdateScreen;
end;

procedure TDpiApplication.Terminate;
begin
  Application.Terminate;
end;

procedure TDpiApplication.ProcessMessages;
begin
  GetNativeApplication.ProcessMessages;
end;

procedure TDpiApplication.UpdateMainForm(AForm: TDpiForm);
begin
  if (FMainForm = nil)
  and (FCreatingForm=AForm)
  //and (not (AppDestroying in FFlags))
  and not (AForm.FormStyle in [fsMDIChild, fsSplash])
  then
    FMainForm := AForm;
  GetNativeApplication.UpdateMainForm(AForm.GetNativeForm);
end;

procedure TDpiApplication.CreateForm(InstanceClass: TComponentClass; out
  Reference);
var
  Instance: TComponent;
  ok: Boolean;
  AForm: TDpiForm;
begin
  // Allocate the instance, without calling the constructor
  Instance := TComponent(InstanceClass.NewInstance);
  // set the Reference before the constructor is called, so that
  // events and constructors can refer to it
  TComponent(Reference) := Instance;

  ok := False;
  try
    if (FCreatingForm = nil) and (Instance is TDpiForm) then
      FCreatingForm := TDpiForm(Instance);
    Instance.Create(Self);
    ok := true;
  finally
    if not ok then begin
      TComponent(Reference) := nil;
      if FCreatingForm = Instance then
        FCreatingForm := nil;
    end;
  end;

  if (Instance is TDpiForm) then begin
    AForm := TDpiForm(Instance);
    UpdateMainForm(AForm);
    if FMainForm = AForm then AForm.GetNativeForm.HandleNeeded;
    if AForm.FormStyle = fsSplash then begin
      // show the splash form and handle the paint message
      AForm.Show;
      AForm.Invalidate;
      ProcessMessages;
    end;
  end;
end;

function TDpiApplication.MessageBox(Text, Caption: PChar; Flags: Longint
  ): Integer;
begin
  Result := Application.MessageBox(Text, Caption, Flags);
end;

{ TDpiJpegImage }

function TDpiJpegImage.GetNativeBitmap: TCustomBitmap;
begin
  Result := GetNativeJpeg;
end;

function TDpiJpegImage.GetNativeJpeg: TJPEGImage;
begin
  Result := NativeJpeg;
end;

constructor TDpiJpegImage.Create;
begin
  NativeJpeg := TJPEGImage.Create;
  inherited;
  NativeGraphicClass := TJPEGImage;
end;

destructor TDpiJpegImage.Destroy;
begin
  FreeAndNil(NativeJpeg);
  inherited Destroy;
end;

{ TDpiPortableNetworkGraphic }

function TDpiPortableNetworkGraphic.GetNativeBitmap: TCustomBitmap;
begin
  Result := GetNativePng;
end;

function TDpiPortableNetworkGraphic.GetNativePng: TPortableNetworkGraphic;
begin
  Result := NativePng;
end;

constructor TDpiPortableNetworkGraphic.Create;
begin
  NativePng := TPortableNetworkGraphic.Create;
  inherited;
  NativeGraphicClass := TPortableNetworkGraphic;
end;

destructor TDpiPortableNetworkGraphic.Destroy;
begin
  Canvas.NativeCanvas := nil;
  FreeAndNil(NativePng);
  inherited;
end;

{ TDpiCustomControl }

function TDpiCustomControl.GetOnPaint: TNotifyEvent;
begin
  Result := GetNativeCustomControl.OnPaint;
end;

function TDpiCustomControl.GetPixelsPerInch: Integer;
begin
  //Result := GetNativeCustomControl.Pix;
end;

function TDpiCustomControl.GetCanvas: TDpiCanvas;
begin
  Result := FCanvas;
end;

procedure TDpiCustomControl.SetOnPaint(AValue: TNotifyEvent);
begin
  GetNativeCustomControl.OnPaint := AValue;
end;

procedure TDpiCustomControl.SetPixelsPerInch(AValue: Integer);
begin

end;

function TDpiCustomControl.GetNativeWinControl: TWinControl;
begin
  Result := GetNativeCustomControl;
end;

function TDpiCustomControl.GetNativeCustomControl: TCustomControl;
begin
  Result := nil;
end;

constructor TDpiCustomControl.Create(TheOwner: TComponent);
begin
  inherited;
  FCanvas := TDpiCanvas.Create;
  FCanvas.NativeCanvas := GetNativeCustomControl.Canvas;
end;

destructor TDpiCustomControl.Destroy;
begin
  FreeAndNil(FCanvas);
  inherited;
end;

{ TDpiScrollBar }

function TDpiScrollBar.GetKind: TScrollBarKind;
begin
  Result := NativeScrollBar.Kind;
end;

function TDpiScrollBar.GetMax: Integer;
begin
  Result := NativeScrollBar.Max;
end;

function TDpiScrollBar.GetMin: Integer;
begin
  Result := NativeScrollBar.Min;
end;

function TDpiScrollBar.GetOnChange: TNotifyEvent;
begin
  Result := NativeScrollBar.OnChange;
end;

function TDpiScrollBar.GetPageSize: Integer;
begin
  Result := NativeScrollBar.PageSize;
end;

function TDpiScrollBar.GetPosition: Integer;
begin
  Result := NativeScrollBar.Position;
end;

procedure TDpiScrollBar.SetKind(AValue: TScrollBarKind);
begin
  NativeScrollBar.Kind := AValue;
end;

procedure TDpiScrollBar.SetMax(AValue: Integer);
begin
  NativeScrollBar.Max := AValue;
end;

procedure TDpiScrollBar.SetMin(AValue: Integer);
begin
  NativeScrollBar.Min := Avalue;
end;

procedure TDpiScrollBar.SetOnChange(AValue: TNotifyEvent);
begin
  NativeScrollBar.OnChange := AValue;
end;

procedure TDpiScrollBar.SetPageSize(AValue: Integer);
begin
  NativeScrollBar.PageSize := AValue;
end;

procedure TDpiScrollBar.SetPosition(AValue: Integer);
begin
  NativeScrollBar.Position := AValue;
end;

function TDpiScrollBar.GetNativeControl: TControl;
begin
  if not Assigned(NativeScrollBar) then NativeScrollBar := TScrollBar.Create(nil);
  Result := NativeScrollBar;
end;

constructor TDpiScrollBar.Create(TheOwner: TComponent);
begin
  inherited;
end;

destructor TDpiScrollBar.Destroy;
begin
  FreeAndNil(NativeScrollBar);
  inherited;
end;

{ TDpiRasterImage }

function TDpiRasterImage.GetRawImage: TRawImage;
begin
  Result := GetNativeRasterImage.RawImage;
end;

function TDpiRasterImage.GetNativeRasterImage: TRasterImage;
begin
  Result := nil;
end;

function TDpiRasterImage.GetNativeGraphic: TGraphic;
begin
  Result := GetNativeRasterImage;
end;

{ TDpiGraphic }

function TDpiGraphic.GetNativeGraphic: TGraphic;
begin
  Result := nil;
end;

procedure TDpiGraphic.ScreenChanged;
begin
end;

procedure TDpiGraphic.SetDpi(AValue: Integer);
begin
  FDpi := AValue;
  ScreenChanged;
end;

function TDpiGraphic.GetDpi: Integer;
begin
  Result := FDpi;
end;

constructor TDpiGraphic.Create;
begin
  Dpi := DpiScreen.Dpi;
end;

procedure StretchDrawBitmap(Src: TRasterImage; Dst: TBitmap);
var
  SrcPtr: TPixelPointer;
  DstPtr: TPixelPointer;
  xx, yy: Integer;
  SrcX, SrcY: Integer;
  DstX, DstY: Integer;
  DstWidth, DstHeight: Integer;
begin
  Dst.BeginUpdate;
  SrcPtr := PixelPointer(Src, 0, 0);
  DstPtr := PixelPointer(Dst, 0, 0);
  {for yy := 0 to Dst.Height - 1 do begin
    for xx := 0 to Dst.Width - 1 do begin
      SrcPtr.SetXY(Min(ScaleFromNative(xx), Src.Width - 1),
        Min(ScaleFromNative(yy), Src.Height - 1));
      DstPtr.Pixel^.B := SrcPtr.Pixel^.B;
      DstPtr.Pixel^.G := SrcPtr.Pixel^.G;
      DstPtr.Pixel^.R := SrcPtr.Pixel^.R;
      DstPtr.NextPixel;
    end;
    DstPtr.NextLine;
  end;
  }
  for SrcY := 0 to Src.Height - 1 do begin
    DstHeight := ScaleToNative(SrcY + 1) - ScaleToNative(SrcY);
    for DstY := 0 to DstHeight - 1 do begin
      for SrcX := 0 to Src.Width - 1 do begin
        DstWidth := ScaleToNative(SrcX + 1) - ScaleToNative(SrcX);
        for DstX := 0 to DstWidth - 1 do begin
          DstPtr.Pixel^.B := SrcPtr.Pixel^.B;
          DstPtr.Pixel^.G := SrcPtr.Pixel^.G;
          DstPtr.Pixel^.R := SrcPtr.Pixel^.R;
          DstPtr.NextPixel;
        end;
        SrcPtr.NextPixel;
      end;
      DstPtr.NextLine;
      SrcPtr.SetX(0);
    end;
    SrcPtr.NextLine;
  end;
  Dst.EndUpdate;
end;

procedure TDpiGraphic.LoadFromFile(const Filename: string);
var
  Bitmap: TGraphic;
begin
  Bitmap := NativeGraphicClass.Create;
  try
    Bitmap.LoadFromFile(FileName);
    Width := Bitmap.Width;
    Height := Bitmap.Height;
    if Self is TDpiBitmap then begin
      StretchDrawBitmap(TRasterImage(Bitmap), TBitmap(GetNativeGraphic));
      //TBitmap(GetNativeGraphic).Canvas.StretchDraw(Bounds(0, 0,
      //TBitmap(GetNativeGraphic).Width, TBitmap(GetNativeGraphic).Height), Bitmap);
    end else raise Exception.Create('Unsupported class ' + Self.ClassName);
  finally
    FreeAndNil(Bitmap);
  end;
end;

procedure TDpiGraphic.SaveToFile(const Filename: string);
var
  Bitmap: TGraphic;
begin
  Bitmap := NativeGraphicClass.Create;
  try
    Bitmap.Width := Width;
    Bitmap.Height := Height;
    if Self is TDpiBitmap then begin
      if Bitmap is TRasterImage then
        (Bitmap as TRasterImage).Canvas.StretchDraw(Bounds(0, 0, Bitmap.Width, Bitmap.Height), TBitmap(GetNativeGraphic))
        else raise Exception.Create('Expected TRasterImage but got ' + Bitmap.ClassName);
    end else raise Exception.Create('Unsupported class ' + Self.ClassName);
    Bitmap.SaveToFile(FileName);
  finally
    FreeAndNil(Bitmap);
  end;
end;

{ TDpiBitmap }

function TDpiBitmap.GetHeight: Integer;
begin
  Result := FHeight;
end;

function TDpiBitmap.GetCanvas: TDpiCanvas;
begin
  Result := FCanvas;
end;

function TDpiBitmap.GetPixelFormat: TPixelFormat;
begin
  Result := GetNativeBitmap.PixelFormat;
end;

function TDpiBitmap.GetScanLine(Row: Integer): Pointer;
begin
  Result := GetNativeBitmap.ScanLine[Row];
end;

function TDpiBitmap.GetWidth: Integer;
begin
  Result := FWidth;
end;

procedure TDpiBitmap.SetHeight(AValue: Integer);
begin
  FHeight := AValue;
  GetNativeBitmap.Height := ScaleToNative(AValue);
end;

procedure TDpiBitmap.SetPixelFormat(AValue: TPixelFormat);
begin
  GetNativeBitmap.PixelFormat := AValue;
end;

procedure TDpiBitmap.SetWidth(AValue: Integer);
begin
  FWidth := AValue;
  GetNativeBitmap.Width := ScaleToNative(AValue);
end;

procedure TDpiBitmap.ScreenChanged;
var
  Bitmap: TBitmap;
  NewWidth: Integer;
  NewHeight: Integer;
begin
  NewWidth := ScaleToNative(Width);
  NewHeight := ScaleToNative(Height);
  if Assigned(NativeBitmap) and ((NewWidth <> NativeBitmap.Width) or (NewHeight <> NativeBitmap.Height)) then begin
    // Rescale bitmap to new size
    Bitmap := TBitmap.Create;
    Bitmap.SetSize(NewWidth, NewHeight);
    Bitmap.PixelFormat := NativeBitmap.PixelFormat;
    Bitmap.Canvas.StretchDraw(Bounds(0, 0, NewWidth, NewHeight), NativeBitmap);
    FreeAndNil(NativeBitmap);
    NativeBitmap := Bitmap;
    Canvas.NativeCanvas := NativeBitmap.Canvas;
  end;
end;

function TDpiBitmap.GetNativeBitmap: TCustomBitmap;
begin
  Result := NativeBitmap;
end;

procedure TDpiBitmap.BeginUpdate;
begin
  GetNativeBitmap.BeginUpdate;
end;

procedure TDpiBitmap.EndUpdate;
begin
  GetNativeBitmap.EndUpdate;
end;

procedure TDpiBitmap.SetSize(AWidth, AHeight: Integer);
begin
  FWidth := AWidth;
  FHeight := AHeight;
  GetNativeBitmap.SetSize(ScaleToNative(AWidth), ScaleToNative(AHeight));
end;

constructor TDpiBitmap.Create;
begin
  inherited;
  NativeGraphicClass := TBitmap;
  NativeBitmap := TBitmap.Create;
  FCanvas := TDpiCanvas.Create;
  FCanvas.NativeCanvas := GetNativeBitmap.Canvas;
end;

destructor TDpiBitmap.Destroy;
begin
  FreeAndNil(FCanvas);
  FreeAndNil(NativeBitmap);
  inherited;
end;

procedure TDpiBitmap.Assign(Source: TPersistent);
begin
  if Source is TDpiBitmap then begin
    GetNativeBitmap.Assign((Source as TDpiBitmap).GetNativeBitmap);
  end else inherited;
end;

function TDpiBitmap.GetNativeRasterImage: TRasterImage;
begin
  Result := GetNativeBitmap;
end;

{ TDpiListBox }

function TDpiListBox.GetBorderStyle: TBorderStyle;
begin
  Result := GetNativeListBox.BorderStyle;
end;

function TDpiListBox.GetCount: Integer;
begin
  Result := GetNativeListBox.Count;
end;

function TDpiListBox.GetExtendedSelect: Boolean;
begin
  Result := GetNativeListBox.ExtendedSelect;
end;

function TDpiListBox.GetIntegralHeight: Boolean;
begin
  Result := GetNativeListBox.IntegralHeight;
end;

function TDpiListBox.GetItemHeight: Integer;
begin
  Result := GetNativeListBox.ItemHeight;
end;

function TDpiListBox.GetItemIndex: Integer;
begin
  Result := GetNativeListBox.ItemIndex;
end;

function TDpiListBox.GetItems: TStrings;
begin
  Result := GetNativeListBox.Items;
end;

function TDpiListBox.GetParentFont: Boolean;
begin
  Result := GetNativeListBox.ParentFont;
end;

function TDpiListBox.GetScrollWidth: Integer;
begin
  Result := GetNativeListBox.ScrollWidth;
end;

function TDpiListBox.GetTopIndex: Integer;
begin
  Result := GetNativeListBox.TopIndex;
end;

procedure TDpiListBox.SetBorderStyle(AValue: TBorderStyle);
begin
  GetNativeListBox.BorderStyle := AValue;
end;

procedure TDpiListBox.SetExtendedSelect(AValue: Boolean);
begin
  GetNativeListBox.ExtendedSelect := AValue;
end;

procedure TDpiListBox.SetIntegralHeight(AValue: Boolean);
begin
  GetNativeListBox.IntegralHeight := AValue;
end;

procedure TDpiListBox.SetItemHeight(AValue: Integer);
begin
  GetNativeListBox.ItemHeight := AValue;
end;

procedure TDpiListBox.SetItemIndex(AValue: Integer);
begin
  GetNativeListBox.ItemIndex := AValue;
end;

procedure TDpiListBox.SetItems(AValue: TStrings);
begin
  GetNativeListBox.Items := AValue;
end;

procedure TDpiListBox.SetParentFont(AValue: Boolean);
begin
  GetNativeListBox.ParentFont := AValue;
end;

procedure TDpiListBox.SetScrollWidth(AValue: Integer);
begin
  GetNativeListBox.ScrollWidth := AValue;
end;

procedure TDpiListBox.SetTopIndex(AValue: Integer);
begin
  GetNativeListBox.TopIndex := AValue;
end;

function TDpiListBox.GetNativeWinControl: TWinControl;
begin
  Result := GetNativeListBox;
end;

function TDpiListBox.GetNativeListBox: TListBox;
begin
  if not Assigned(NativeListBox) then NativeListBox := TListBox.Create(nil);
  Result := NativeListBox;
end;

destructor TDpiListBox.Destroy;
begin
  FreeAndNil(NativeListBox);
  inherited Destroy;
end;

{ TDpiPaintBox }

function TDpiPaintBox.GetNativeGraphicControl: TGraphicControl;
begin
  if not Assigned(NativePaintBox) then NativePaintBox := TPaintBox.Create(nil);
  Result := NativePaintBox;
end;

constructor TDpiPaintBox.Create(TheOwner: TComponent);
begin
  inherited;
  Canvas := TDpiCanvas.Create;
  Canvas.NativeCanvas := NativePaintBox.Canvas;
  Canvas.Font.NativeFont := NativePaintBox.Canvas.Font;
  UpdateNativeControl;
  ScreenChanged;
end;

destructor TDpiPaintBox.Destroy;
begin
  FreeAndNil(NativePaintBox);
  inherited;
end;

{ TDpiPicture }

procedure TDpiPicture.SetBitmap(AValue: TDpiBitmap);
begin
  if FBitmap = AValue then Exit;
  FBitmap := AValue;
end;

procedure TDpiPicture.LoadFromFile(FileName: string);
begin
end;

{ TDpiCanvas }

function TDpiCanvas.GetBrush: TBrush;
begin
  Result := GetNativeCanvas.Brush;
end;

function TDpiCanvas.GetHandle: HDC;
begin
  Result := GetNativeCanvas.Handle;
end;

function TDpiCanvas.GetHeight: Integer;
begin
  Result := ScaleFromNative(GetNativeCanvas.Height);
end;

function TDpiCanvas.GetPen: TPen;
begin
  Result := GetNativeCanvas.Pen;
end;

function TDpiCanvas.GetPixel(X, Y: Integer): TColor;
begin
  Result := GetNativeCanvas.Pixels[ScaleToNative(X), ScaleToNative(Y)];
end;

function TDpiCanvas.GetWidth: Integer;
begin
  Result := ScaleFromNative(GetNativeCanvas.Width);
end;

procedure TDpiCanvas.SetBrush(AValue: TBrush);
begin
  GetNativeCanvas.Brush := AValue;
end;

procedure TDpiCanvas.SetFont(AValue: TDpiFont);
begin
  if FFont = AValue then Exit;
  if FFontFree then FreeAndNil(FFont);
  FFontFree := False;
  FFont := AValue;
end;

procedure TDpiCanvas.SetHandle(AValue: HDC);
begin
  GetNativeCanvas.Handle := AValue;
end;

procedure TDpiCanvas.SetPen(AValue: TPen);
begin
  GetNativeCanvas.Pen := AValue;
end;

procedure TDpiCanvas.SetPixel(X, Y: Integer; AValue: TColor);
var
  BrushStyle: TBrushStyle;
  BrushColor: TColor;
begin
 { BrushStyle := GetNativeCanvas.Brush.Style;
  BrushColor := GetNativeCanvas.Brush.Color;
  GetNativeCanvas.Brush.Color := AValue;
  GetNativeCanvas.Brush.Style := bsClear;
  GetNativeCanvas.FillRect(ScaleToNative(X), ScaleToNative(Y), ScaleToNative(X + 1) - 1, ScaleToNative(Y + 1) - 1);
  GetNativeCanvas.Brush.Style := BrushStyle;
  GetNativeCanvas.Brush.Color := BrushColor;
  }
  GetNativeCanvas.Pixels[ScaleToNative(X), ScaleToNative(Y)] := AValue;
end;

procedure TDpiCanvas.SetNativeCanvas(AValue: TCanvas);
begin
  if FNativeCanvas = AValue then Exit;
  if FNativeCanvasFree then FreeAndNil(FNativeCanvas);
  FNativeCanvasFree := False;
  FNativeCanvas := AValue;
  if Assigned(FNativeCanvas) then begin
    FFont.NativeFont := FNativeCanvas.Font;
  end;
end;

function TDpiCanvas.GetNativeCanvas: TCanvas;
begin
  Result := NativeCanvas;
end;

procedure TDpiCanvas.StretchDraw(const DestRect: TRect; SrcGraphic: TDpiGraphic
  );
begin
  GetNativeCanvas.StretchDraw(ScaleRectToNative(DestRect), SrcGraphic.GetNativeGraphic);
end;

procedure TDpiCanvas.FrameRect(Rect: TRect);
begin
  GetNativeCanvas.FrameRect(ScaleRectToNative(Rect));
end;

procedure TDpiCanvas.Rectangle(X1, Y1, X2, Y2: Integer);
begin
  GetNativeCanvas.Rectangle(ScaleToNative(X1), ScaleToNative(Y1), ScaleToNative(X2), ScaleToNative(Y2));
end;

procedure TDpiCanvas.Rectangle(const ARect: TRect);
begin
  Rectangle(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
end;

function TDpiCanvas.TextWidth(Text: string): Integer;
begin
  Result := ScaleFromNative(GetNativeCanvas.TextWidth(Text));
end;

function TDpiCanvas.TextHeight(Text: string): Integer;
begin
  Result := ScaleFromNative(GetNativeCanvas.TextHeight(Text));
end;

function TDpiCanvas.TextExtent(Text: string): TSize;
begin
  Result := ScaleSizeFromNative(GetNativeCanvas.TextExtent(Text));
end;

procedure TDpiCanvas.TextOut(X, Y: Integer; Text: string);
begin
  GetNativeCanvas.TextOut(ScaleToNative(X), ScaleToNative(Y), Text);
end;

procedure TDpiCanvas.TextRect(ARect: TRect; X, Y: Integer; Text: string);
begin
  GetNativeCanvas.TextRect(ScaleRectToNative(ARect), ScaleToNative(X), ScaleToNative(Y), Text);
end;

procedure TDpiCanvas.MoveTo(X, Y: Integer);
begin
  GetNativeCanvas.MoveTo(ScaleToNative(X), ScaleToNative(Y));
end;

procedure TDpiCanvas.LineTo(X, Y: Integer);
begin
  GetNativeCanvas.LineTo(ScaleToNative(X), ScaleToNative(Y));
end;

procedure TDpiCanvas.FillRect(ARect: TRect);
begin
  GetNativeCanvas.FillRect(ScaleRectToNative(ARect));
end;

procedure TDpiCanvas.FillRect(X1, Y1, X2, Y2: Integer);
begin
  GetNativeCanvas.FillRect(ScaleToNative(X1), ScaleToNative(Y1), ScaleToNative(X2), ScaleToNative(Y2));
end;

procedure TDpiCanvas.Draw(X, Y: Integer; Source: TDpiGraphic);
begin
  GetNativeCanvas.Draw(ScaleToNative(X), ScaleToNative(Y), Source.GetNativeGraphic);
end;

procedure TDpiCanvas.CopyRect(Dest: TRect; SrcCanvas: TDpiCanvas;
  Source: TRect);
begin
  GetNativeCanvas.CopyRect(Dest, SrcCanvas.NativeCanvas, ScaleRectToNative(Source));
end;

constructor TDpiCanvas.Create;
begin
  FNativeCanvas := nil;
  FFont := TDpiFont.Create;
  FFontFree := True;
end;

destructor TDpiCanvas.Destroy;
begin
  if FFontFree then FreeAndNil(FFont);
  if FNativeCanvasFree then FreeAndNil(FNativeCanvasFree);
  inherited;
end;

{ TDpiGraphicControl }

procedure TDpiGraphicControl.SetCanvas(AValue: TDpiCanvas);
begin
  if FCanvas = AValue then Exit;
  FCanvas := AValue;
end;

procedure TDpiGraphicControl.PaintHandler(Sender: TObject);
begin
  Paint;
  if Assigned(FOnPaint) then
    FOnPaint(Sender);
end;

procedure TDpiGraphicControl.Paint;
begin
end;

function TDpiGraphicControl.GetNativeControl: TControl;
begin
  Result := GetNativeGraphicControl;
end;

function TDpiGraphicControl.GetNativeGraphicControl: TGraphicControl;
begin
  Result := NativeGraphicControl;
end;

procedure TDpiGraphicControl.UpdateNativeControl;
begin
  inherited;
  TGraphicControlEx(GetNativeGraphicControl).OnPaint := @PaintHandler;
end;

function TDpiGraphicControl.GetOnPaint: TNotifyEvent;
begin
  Result := FOnPaint;
end;

procedure TDpiGraphicControl.SetOnPaint(AValue: TNotifyEvent);
begin
  FOnPaint := AValue;
end;

constructor TDpiGraphicControl.Create(TheOwner: TComponent);
begin
  NativeGraphicControl := TGraphicControl.Create(nil);
  inherited;
  FCanvas := TDpiCanvas.Create;
  FCanvas.NativeCanvas := GetNativeGraphicControl.Canvas;
end;

destructor TDpiGraphicControl.Destroy;
begin
  FreeAndNil(FCanvas);
  FreeAndNil(NativeGraphicControl);
  inherited;
end;


{ TDpiImage }

procedure TDpiImage.SetStretch(AValue: Boolean);
begin
  if FStretch = AValue then Exit;
  FStretch := AValue;
  NativeImage.Stretch := AValue;
end;

procedure TDpiImage.SetPicture(AValue: TDpiPicture);
begin
  if FDpiPicture = AValue then Exit;
  FDpiPicture := AValue;
end;

function TDpiImage.GetNativeControl: TControl;
begin
  if not Assigned(NativeImage) then NativeImage := TImage.Create(nil);
  Result := NativeImage;
end;

destructor TDpiImage.Destroy;
begin
  FreeAndNil(NativeImage);
  inherited Destroy;
end;

{ TDpiFont }

procedure TDpiFont.SetSize(AValue: Integer);
begin
  if FSize = AValue then Exit;
  FSize := AValue;
  FHeight := -MulDiv(FSize, FPixelsPerInch, 72);
  UpdateFont;
  DoChange;
end;

procedure TDpiFont.DoChange;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDpiFont.SetStyle(AValue: TFontStyles);
begin
  GetNativeFont.Style := AValue;
end;

procedure TDpiFont.UpdateFont;
begin
  if Assigned(GetNativeFont) then begin
    GetNativeFont.PixelsPerInch := FPixelsPerInch;
    GetNativeFont.Size := FSize;
  end;
end;

procedure TDpiFont.ScreenChanged;
begin
  DoChange;
end;

function TDpiFont.GetNativeFont: TFont;
begin
  Result := NativeFont;
end;

procedure TDpiFont.SetOnChange(AValue: TNotifyEvent);
begin
  if FOnChange = AValue then Exit;
  FOnChange := AValue;
end;

procedure TDpiFont.SetPixelsPerInch(AValue: Integer);
begin
  FPixelsPerInch := PixelsPerInch;
  FHeight := -MulDiv(FSize, FPixelsPerInch, 72);
  UpdateFont;
end;

function TDpiFont.GetName: string;
begin
  Result := GetNativeFont.Name;
end;

function TDpiFont.GetColor: TColor;
begin
  Result := FColor;
end;

function TDpiFont.GetCharSet: TFontCharSet;
begin
  Result := GetNativeFont.CharSet;
end;

function TDpiFont.GetHeight: Integer;
begin
  Result := GetNativeFont.Height;
end;

function TDpiFont.GetPixelsPerInch: Integer;
begin
  Result := FPixelsPerInch;
end;

function TDpiFont.GetSize: Integer;
begin
  Result := FSize;
end;

function TDpiFont.GetStyle: TFontStyles;
begin
  Result := GetNativeFont.Style;
end;

function TDpiFont.IsNameStored: Boolean;
begin
  Result := GetNativeFont.Name <> 'default';
end;

procedure TDpiFont.SetCharSet(AValue: TFontCharSet);
begin
  GetNativeFont.CharSet := AValue;
end;

procedure TDpiFont.SetColor(AValue: TColor);
begin
  if FColor = AValue then Exit;
  FColor := AValue;
  GetNativeFont.Color := AValue;
end;

procedure TDpiFont.SetHeight(AValue: Integer);
begin
  FHeight := AValue;
  FSize := MulDiv(-FHeight, 72, FPixelsPerInch);
  UpdateFont;
end;

procedure TDpiFont.SetName(AValue: string);
begin
  GetNativeFont.Name := AValue;
end;

procedure TDpiFont.SetNativeFont(AValue: TFont);
begin
  if FNativeFont = AValue then Exit;
  if FNativeFontFree then FreeAndNil(FNativeFont);
  FNativeFontFree := False;
  FNativeFont := AValue;
end;

constructor TDpiFont.Create;
begin
  FNativeFont := TFont.Create;
  FNativeFontFree := True;
  FPixelsPerInch := DpiScreen.PixelsPerInch;
  Size := 8;
end;

destructor TDpiFont.Destroy;
begin
  if FNativeFontFree then
    FreeAndNil(FNativeFont);
  inherited;
end;

procedure TDpiFont.Assign(Source: TPersistent);
begin
  if Source is TDpiFont then begin
    GetNativeFont.Assign((Source as TDpiFont).GetNativeFont);
    Size := (Source as TDpiFont).Size;
    Height := (Source as TDpiFont).Height;
    PixelsPerInch := (Source as TDpiFont).PixelsPerInch;
    FOnChange := (Source as TDpiFont).FOnChange;
    Color := (Source as TDpiFont).Color;
  end;
end;

{ TDpiWinControl }

function TDpiWinControl.GetBorderStyle: TBorderStyle;
begin
  Result := TWinControlEx(GetNativeWinControl).BorderStyle;
end;

function TDpiWinControl.GetHandle: HWND;
begin
  Result := GetNativeWinControl.Handle;
end;

function TDpiWinControl.GetOnKeyDown: TKeyEvent;
begin
  Result := GetNativeWinControl.OnKeyDown;
end;

function TDpiWinControl.GetOnKeyPress: TKeyPressEvent;
begin
  Result := GetNativeWinControl.OnKeyPress;
end;

function TDpiWinControl.GetOnKeyUp: TKeyEvent;
begin
  Result := GetNativeWinControl.OnKeyUp;
end;

function TDpiWinControl.GetTabOrder: TTabOrder;
begin
  Result := GetNativeWinControl.TabOrder;
end;

function TDpiWinControl.GetTabStop: Boolean;
begin
  Result := GetNativeWinControl.TabStop;
end;

procedure TDpiWinControl.SetBorderStyle(AValue: TBorderStyle);
begin
  TWinControlEx(GetNativeWinControl).BorderStyle := AValue;
end;

procedure TDpiWinControl.SetHandle(AValue: HWND);
begin
  GetNativeWinControl.Handle := AValue;
end;

procedure TDpiWinControl.SetOnKeyDown(AValue: TKeyEvent);
begin
  GetNativeWinControl.OnKeyDown := AValue;
end;

procedure TDpiWinControl.SetOnKeyPress(AValue: TKeyPressEvent);
begin
  GetNativeWinControl.OnKeyPress := AValue;
end;

procedure TDpiWinControl.SetOnKeyUp(AValue: TKeyEvent);
begin
  GetNativeWinControl.OnKeyUp := AValue;
end;

procedure TDpiWinControl.SetTabOrder(AValue: TTabOrder);
begin
  GetNativeWinControl.TabOrder := AValue;
end;

procedure TDpiWinControl.SetTabStop(AValue: Boolean);
begin
  GetNativeWinControl.TabStop := AValue;
end;

function TDpiWinControl.GetNativeControl: TControl;
begin
  Result := GetNativeWinControl;
end;

function TDpiWinControl.GetNativeWinControl: TWinControl;
begin
  Result := nil;
end;

procedure TDpiWinControl.ScreenChanged;
var
  I: Integer;
begin
  inherited;
  for I := 0 to Controls.Count - 1 do
    Controls[I].ScreenChanged;
end;

function TDpiWinControl.ControlCount: Integer;
begin
  Result := Controls.Count;
end;

constructor TDpiWinControl.Create(TheOwner: TComponent);
begin
  Controls := TDpiControls.Create;
  Controls.FreeObjects := False;
  inherited;
end;

destructor TDpiWinControl.Destroy;
begin
  FreeAndNil(Controls);
  inherited;
end;

{ TDpiScreen }

procedure TDpiScreen.SetDpi(AValue: Integer);
begin
  if FDpi = AValue then Exit;
  FDpi := AValue;
  UpdateForms;
end;

function TDpiScreen.GetWidth: Integer;
begin
  Result := ScaleFromNative(Screen.Width);
end;

procedure TDpiScreen.SetCursor(AValue: TCursor);
begin
  Screen.Cursor := AValue;
end;

procedure TDpiScreen.SetCursors(Index: Integer; AValue: HCURSOR);
begin
  Screen.Cursors[Index] := AValue;
end;

function TDpiScreen.GetHeight: Integer;
begin
  Result := ScaleFromNative(Screen.Height);
end;

function TDpiScreen.GetFormCount: Integer;
begin
  Result := FForms.Count;
end;

function TDpiScreen.GetForms(Index: Integer): TDpiForm;
begin
  Result := FForms[Index];
end;

procedure TDpiScreen.AddForm(AForm: TDpiForm);
begin
  if AForm is TDpiForm then begin
    FForms.Add(AForm);
    //DpiApplication.UpdateVisible;
  end;
end;

procedure TDpiScreen.RemoveForm(AForm: TDpiForm);
begin
  FForms.Remove(AForm);
  if AForm = FActiveForm then begin
    FActiveFormPrev := FActiveForm;
    FActiveForm := nil;
  end;
end;

function TDpiScreen.GetActiveForm: TDpiForm;
begin
  Result := FActiveForm;
end;

function TDpiScreen.GetCursor: TCursor;
begin
  Result := Screen.Cursor;
end;

function TDpiScreen.GetCursors(Index: Integer): HCURSOR;
begin
  Result := Screen.Cursors[Index];
end;

procedure TDpiScreen.UpdateForms;
var
  I: Integer;
begin
  for I := 0 to FForms.Count - 1 do
    FForms[I].ScreenChanged;
end;

constructor TDpiScreen.Create;
begin
  FForms := TDpiForms.Create;
  FForms.FreeObjects := False;
  FPrevActiveForms := TDpiForms.Create;
  FPrevActiveForms.FreeObjects := False;
  // Screen.PixelsPerInch is not initialized at this point
  Dpi := 96;
end;

destructor TDpiScreen.Destroy;
begin
  FreeAndNil(FForms);
  FreeAndNil(FPrevActiveForms);
  inherited Destroy;
end;

procedure TDpiScreen.UpdateScreen;
begin
  Dpi := Screen.PixelsPerInch;
end;

procedure TDpiScreen.UpdateActiveFormFromNativeScreen;
var
  I: Integer;
  F: TDpiForm;
begin
  if Screen.ActiveForm = nil then FActiveForm := nil
  else begin
    for I := 0 to FormCount - 1 do begin
      F := Forms[I];
      if F.GetNativeForm = Screen.ActiveForm then begin
        FActiveForm := F;
        Break;
      end;
    end;
  end;
end;

{ TDpiButton }

function TDpiButton.GetNativeControl: TControl;
begin
  if not Assigned(NativeButton) then NativeButton := TButton.Create(nil);
  Result := NativeButton;
end;

destructor TDpiButton.Destroy;
begin
  FreeAndNil(NativeButton);
  inherited;
end;

{ TDpiControl }

procedure TDpiControl.SetTop(AValue: Integer);
begin
  if FTop = AValue then Exit;
  FTop := AValue;
  UpdateBounds;
end;

procedure TDpiControl.SetVisible(AValue: Boolean);
begin
  GetNativeControl.Visible := AValue;
end;

procedure TDpiControl.SetWidth(AValue: Integer);
begin
  if FWidth = AValue then Exit;
  FWidth := AValue;
  UpdateBounds;
end;

function TDpiControl.GetNativeControl: TControl;
begin
  Result := nil;
end;

procedure TDpiControl.UpdateNativeControl;
begin
  Font.NativeFont := GetNativeControl.Font;
  GetNativeControl.OnResize := @NativeFormResize;
  GetNativeControl.OnChangeBounds := @NativeChangeBounds;
  TControlEx(GetNativeControl).OnMouseDown := @MouseDownHandler;
  TControlEx(GetNativeControl).OnMouseUp := @MouseUpHandler;
  TControlEx(GetNativeControl).OnMouseMove := @MouseMoveHandler;
  TControlEx(GetNativeControl).OnMouseEnter := @MouseEnterHandler;
  TControlEx(GetNativeControl).OnMouseLeave := @MouseLeaveHandler;
  TControlEx(GetNativeControl).OnMouseWheel := @MouseWheelHandler;
end;

procedure TDpiControl.MouseDownHandler(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseDown(Button, Shift, ScaleFromNative(X), ScaleFromNative(Y));
  if Assigned(FOnMouseDown) then FOnMouseDown(Self, Button, Shift, ScaleFromNative(X), ScaleFromNative(Y));
end;

procedure TDpiControl.MouseUpHandler(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseUp(Button, Shift, ScaleFromNative(X), ScaleFromNative(Y));
  if Assigned(FOnMouseUp) then FOnMouseUp(Self, Button, Shift, ScaleFromNative(X), ScaleFromNative(Y));
end;

procedure TDpiControl.MouseMoveHandler(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  MouseMove(Shift, ScaleFromNative(X), ScaleFromNative(Y));
  if Assigned(FOnMouseMove) then FOnMouseMove(Self, Shift, ScaleFromNative(X), ScaleFromNative(Y));
end;

procedure TDpiControl.MouseWheelHandler(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  if Assigned(FOnMouseWheel) then FOnMouseWheel(Self, Shift, WheelDelta,
    ScalePointFromNative(MousePos), Handled);
end;

procedure TDpiControl.MouseLeaveHandler(Sender: TObject);
begin
  MouseLeave;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

procedure TDpiControl.MouseEnterHandler(Sender: TObject);
begin
  MouseEnter;
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TDpiControl.DoBorderSpacingChange(Sender: TObject;
  InnerSpaceChanged: Boolean);
begin
  GetNativeControl.BorderSpacing.Left := ScaleToNative(FBorderSpacing.Left);
  GetNativeControl.BorderSpacing.Right := ScaleToNative(FBorderSpacing.Right);
  GetNativeControl.BorderSpacing.Top := ScaleToNative(FBorderSpacing.Top);
  GetNativeControl.BorderSpacing.Bottom := ScaleToNative(FBorderSpacing.Bottom);
  GetNativeControl.BorderSpacing.Around := ScaleToNative(FBorderSpacing.Around);
end;

procedure TDpiControl.SetText(AValue: TCaption);
begin

end;

procedure TDpiControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
end;

procedure TDpiControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
end;

procedure TDpiControl.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
end;

procedure TDpiControl.MouseLeave;
begin
end;

procedure TDpiControl.MouseEnter;
begin
end;

function TDpiControl.ScreenToClient(const APoint: TPoint): TPoint;
begin
  Result := ScalePointFromNative(GetNativeControl.ScreenToClient(ScalePointToNative(APoint)));
end;

function TDpiControl.ClientToScreen(const APoint: TPoint): TPoint;
begin
  Result := ScalePointFromNative(GetNativeControl.ClientToScreen(ScalePointToNative(APoint)));
end;

procedure TDpiControl.AddHandlerOnVisibleChanged(
  const OnVisibleChangedEvent: TNotifyEvent; AsFirst: boolean);
begin
  GetNativeControl.AddHandlerOnVisibleChanged(OnVisibleChangedEvent, AsFirst);
end;

procedure TDpiControl.RemoveHandlerOnVisibleChanged(
  const OnVisibleChangedEvent: TNotifyEvent);
begin
  GetNativeControl.RemoveHandlerOnVisibleChanged(OnVisibleChangedEvent);
end;

procedure TDpiControl.ScreenChanged;
begin
  UpdateBounds;
  Font.ScreenChanged;
end;

procedure TDpiControl.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  FLeft := ALeft;
  FTop := ATop;
  FWidth := AWidth;
  FHeight := AHeight;
  UpdateBounds;
end;

procedure TDpiControl.Show;
begin
  Visible := True;
end;

procedure TDpiControl.Hide;
begin
  Visible := False;
end;

procedure TDpiControl.Invalidate;
begin
  GetNativeControl.Invalidate;
end;

procedure TDpiControl.Repaint;
begin
  GetNativeControl.Repaint;
end;

procedure TDpiControl.Update;
begin
  GetNativeControl.Update;
end;

function TDpiControl.IsParentOf(AControl: TDpiControl): boolean;
begin
  Result := False;
  while Assigned(AControl) do
  begin
    AControl := AControl.Parent;
    if Self = AControl then
      Exit(True);
  end;
end;

constructor TDpiControl.Create(TheOwner: TComponent);
begin
  inherited;
  FFont := TDpiFont.Create;
  FFont.OnChange := @FontChanged;
  FConstraints := TDpiSizeConstraints.Create;
  if Assigned(TheOwner) and (TheOwner is TDpiWinControl) then
    Parent := TDpiWinControl(TheOwner);
  FBorderSpacing := TDpiControlBorderSpacing.Create(Self);
  GetNativeControl;
  UpdateNativeControl;
  ScreenChanged;
end;

destructor TDpiControl.Destroy;
begin
  FreeAndNil(FBorderSpacing);
  FreeAndNil(FConstraints);
  FreeAndNil(FFont);
  inherited;
end;

procedure TDpiControl.SetLeft(AValue: Integer);
begin
  if FLeft = AValue then Exit;
  FLeft := AValue;
  UpdateBounds;
end;

procedure TDpiControl.SetCaption(AValue: string);
begin
  GetNativeControl.Caption := AValue;
end;

procedure TDpiControl.SetParent(AValue: TDpiWinControl);
begin
  if FParent = AValue then Exit;
  if Assigned(FParent) then begin
    FParent.Controls.Remove(Self);
    if Assigned(FParent) and (FParent is TDpiWinControl) then
      GetNativeControl.Parent := nil;
  end;
  FParent := AValue;
  if Assigned(FParent) then begin
    FParent.Controls.Add(Self);
    if Assigned(FParent) and (FParent is TDpiWinControl) then
      GetNativeControl.Parent := TDpiWinControl(FParent).GetNativeWinControl;
  end;
end;

procedure TDpiControl.SetFont(AValue: TDpiFont);
begin
  if FFont = AValue then Exit;
  FFont := AValue;
end;

procedure TDpiControl.SetHint(AValue: string);
begin
  GetNativeControl.Hint := AValue;
end;

function TDpiControl.GetBoundsRect: TRect;
begin
  Result.Left := Left;
  Result.Top := Top;
  Result.Right := Left + Width;
  Result.Bottom := Top + Height;
end;

function TDpiControl.GetAlign: TAlign;
begin
  Result := GetNativeControl.Align;
end;

function TDpiControl.GetAnchors: TAnchors;
begin
  Result := GetNativeControl.Anchors;
end;

function TDpiControl.GetClientHeight: Integer;
begin
  Result := ScaleFromNative(GetNativeControl.ClientHeight);
end;

function TDpiControl.GetClientWidth: Integer;
begin
  Result := ScaleFromNative(GetNativeControl.ClientWidth);
end;

function TDpiControl.GetColor: TColor;
begin
  Result := GetNativeControl.Color;
end;

function TDpiControl.GetCursor: TCursor;
begin
  Result := GetNativeControl.Cursor;
end;

function TDpiControl.GetEnabled: Boolean;
begin
  Result := GetNativeControl.Enabled;
end;

function TDpiControl.GetHint: string;
begin
  Result := GetNativeControl.Hint;
end;

function TDpiControl.GetOnClick: TNotifyEvent;
begin
  Result := GetNativeControl.OnClick;
end;

function TDpiControl.GetParentFont: Boolean;
begin
  Result := TControlEx(GetNativeControl).ParentFont;
end;

function TDpiControl.GetShowHint: Boolean;
begin
  Result := GetNativeControl.ShowHint;
end;

function TDpiControl.GetText: TCaption;
begin
  Result := '';
end;

function TDpiControl.GetVisible: Boolean;
begin
  Result := GetNativeControl.Visible;
end;

function TDpiControl.IsAnchorsStored: Boolean;
begin

end;

procedure TDpiControl.SetAlign(AValue: TAlign);
begin
  GetNativeControl.Align := AValue;
end;

procedure TDpiControl.SetAnchors(AValue: TAnchors);
begin
  GetNativeControl.Anchors := AValue;
end;

procedure TDpiControl.SetBorderSpacing(AValue: TDpiControlBorderSpacing);
begin
  if FBorderSpacing = AValue then Exit;
  FBorderSpacing := AValue;
end;

procedure TDpiControl.SetBoundsRect(AValue: TRect);
begin
  SetBounds(AValue.Left, AValue.Top, AValue.Right - AValue.Left, AValue.Bottom - AValue.Top);
end;

procedure TDpiControl.SetClientHeight(AValue: Integer);
begin
  GetNativeControl.ClientHeight := ScaleToNative(AValue);
end;

procedure TDpiControl.SetClientWidth(AValue: Integer);
begin
  GetNativeControl.ClientWidth := ScaleToNative(AValue);
end;

procedure TDpiControl.SetColor(AValue: TColor);
begin
  GetNativeControl.Color := AValue;
end;

procedure TDpiControl.SetCursor(AValue: TCursor);
begin
  GetNativeControl.Cursor := AValue;
end;

procedure TDpiControl.SetEnabled(AValue: Boolean);
begin
  GetNativeControl.Enabled := AValue;
end;

procedure TDpiControl.SetOnChangeBounds(AValue: TNotifyEvent);
begin
  if FOnChangeBounds = AValue then Exit;
  FOnChangeBounds := AValue;
end;

procedure TDpiControl.SetOnClick(AValue: TNotifyEvent);
begin
  GetNativeControl.OnClick := AValue;
end;

procedure TDpiControl.SetOnResize(AValue: TNotifyEvent);
begin
  if FOnResize = AValue then Exit;
  FOnResize := AValue;
end;

procedure TDpiControl.SetParentFont(AValue: Boolean);
begin
  TControlEx(GetNativeControl).ParentFont := AValue;
end;

procedure TDpiControl.SetShowHint(AValue: Boolean);
begin
  GetNativeControl.ShowHint := AValue;
end;

procedure TDpiControl.NativeFormResize(Sender: TObject);
begin
  BoundsRect := ScaleRectFromNative(GetNativeControl.BoundsRect);
  DoFormResize;
end;

procedure TDpiControl.NativeChangeBounds(Sender: TObject);
var
  NewBounds: TRect;
begin
  NewBounds := ScaleRectFromNative(GetNativeControl.BoundsRect);
  if NewBounds <> BoundsRect then begin
    FLeft := NewBounds.Left;
    FTop := NewBounds.Top;
    FWidth := NewBounds.Width;
    FHeight := NewBounds.Height;
    DoChangeBounds;
  end;
end;

procedure TDpiControl.DoFormResize;
begin
  if Assigned(FOnResize) then begin
    FOnResize(Self);
  end;
end;

procedure TDpiControl.DoChangeBounds;
begin
  if Assigned(FOnChangeBounds) then FOnChangeBounds(Self);
end;

function TDpiControl.GetCaption: string;
begin
  Result := GetNativeControl.Caption;
end;

procedure TDpiControl.FontChanged(Sender: TObject);
begin
  GetNativeControl.Font.Size := ScaleToNative(Font.Size);
end;

procedure TDpiControl.UpdateBounds;
var
  R: TRect;
begin
  GetNativeControl.BoundsRect := ScaleRectToNative(BoundsRect);
  R := ScaleRectToNative(BoundsRect);
  //WriteLog(Name + ' ' + IntToStr(R.Left) + ', ' + IntToStr(R.Top) + ', ' + IntToStr(R.Width) + ', ' + IntToStr(R.Height))
end;

procedure TDpiControl.SetHeight(AValue: Integer);
begin
  if FHeight = AValue then Exit;
  FHeight := AValue;
  UpdateBounds;
end;

{ TDpiFormFileDesc }

constructor TDpiFormFileDesc.Create;
begin
  inherited Create;
  Name := 'DpiForm form';
  ResourceClass := TDpiForm;
  UseCreateFormStatements := False;
end;

function TDpiFormFileDesc.GetInterfaceUsesSection: string;
begin
  Result := inherited GetInterfaceUsesSection;
  Result := Result + ', UDpiForm';
end;

function TDpiFormFileDesc.GetLocalizedName: string;
begin
  Result := SDpiFormTitle;
end;

function TDpiFormFileDesc.GetLocalizedDescription: string;
begin
  Result := SDpiFormDescription;
end;

{ TDpiForm }

function TDpiForm.GetBorderIcons: TBorderIcons;
begin
  Result := GetNativeForm.BorderIcons;
end;

function TDpiForm.GetBorderStyle: TFormBorderStyle;
begin
  Result := GetNativeForm.BorderStyle;
end;

function TDpiForm.GetDesignTimePPI: Integer;
begin
  Result := GetNativeForm.DesignTimePPI;
end;

function TDpiForm.GetFormState: TFormState;
begin
  Result := GetNativeForm.FormState;
end;

function TDpiForm.GetFormStyle: TFormStyle;
begin
  Result := GetNativeForm.FormStyle;
end;

function TDpiForm.GetKeyPreview: Boolean;
begin
  Result := GetNativeForm.KeyPreview;
end;

function TDpiForm.GetLCLVersion: string;
begin
  Result := GetNativeForm.LCLVersion;
end;

function TDpiForm.GetModalResult: TModalResult;
begin
  Result := GetNativeForm.ModalResult;
end;

function TDpiForm.GetOnCloseQuery: TCloseQueryEvent;
begin
  Result := GetNativeForm.OnCloseQuery;
end;

function TDpiForm.GetOnCreate: TNotifyEvent;
begin
  Result := GetNativeForm.OnCreate;
end;

function TDpiForm.GetOnDeactivate: TNotifyEvent;
begin
  Result := GetNativeForm.OnDeactivate;
end;

function TDpiForm.GetOnDestroy: TNotifyEvent;
begin
  Result := GetNativeForm.OnDestroy;
end;

function TDpiForm.GetOnHide: TNotifyEvent;
begin
  Result := GetNativeForm.OnHide;
end;

function TDpiForm.GetOnShow: TNotifyEvent;
begin
  Result := GetNativeForm.OnShow;
end;

function TDpiForm.GetPosition: TPosition;
begin
  Result := GetNativeForm.Position;
end;

function TDpiForm.GetWindowState: TWindowState;
begin
  Result := GetNativeForm.WindowState;
end;

procedure TDpiForm.SetBorderIcons(AValue: TBorderIcons);
begin
  GetNativeForm.BorderIcons := AValue;
end;

procedure TDpiForm.SetBorderStyle(AValue: TFormBorderStyle);
begin
  GetNativeForm.BorderStyle := AValue;
end;

procedure TDpiForm.SetDesignTimePPI(AValue: Integer);
begin
  GetNativeForm.DesignTimePPI := AValue;
end;

procedure TDpiForm.SetFormStyle(AValue: TFormStyle);
begin
  GetNativeForm.FormStyle := AValue;
end;

procedure TDpiForm.SetKeyPreview(AValue: Boolean);
begin
  GetNativeForm.KeyPreview := AValue;
end;

procedure TDpiForm.SetLCLVersion(AValue: string);
begin
  GetNativeForm.LCLVersion := AValue;
end;

procedure TDpiForm.SetModalResult(AValue: TModalResult);
begin
  GetNativeForm.ModalResult := AValue;
end;

procedure TDpiForm.SetOnCloseQuery(AValue: TCloseQueryEvent);
begin
  GetNativeForm.OnCloseQuery := AValue;
end;

procedure TDpiForm.SetOnCreate(AValue: TNotifyEvent);
begin
  GetNativeForm.OnCreate := AValue;
end;

procedure TDpiForm.SetOnDeactivate(AValue: TNotifyEvent);
begin
  GetNativeForm.OnDeactivate := AValue;
end;

procedure TDpiForm.SetOnDestroy(AValue: TNotifyEvent);
begin
  GetNativeForm.OnDestroy := AValue;
end;

procedure TDpiForm.SetOnHide(AValue: TNotifyEvent);
begin
  GetNativeForm.OnHide := AValue;
end;

procedure TDpiForm.SetOnShow(AValue: TNotifyEvent);
begin
  GetNativeForm.OnShow := AValue;
end;

procedure TDpiForm.DoOnCreate;
begin
  if Assigned(GetNativeForm.OnCreate) then
    GetNativeForm.OnCreate(Self);
end;

procedure TDpiForm.FormMessageHandler(var TheMessage: TLMessage);
begin
  Dispatch(TheMessage);
end;

procedure TDpiForm.SetPosition(AValue: TPosition);
begin
  GetNativeForm.Position := AValue;
end;

procedure TDpiForm.SetWindowState(AValue: TWindowState);
begin
  GetNativeForm.WindowState := AValue;
end;

procedure TDpiForm.ActivateHandler(Sender: TObject);
begin
  DpiScreen.FActiveFormPrev := DpiScreen.FActiveForm;
  DpiScreen.FActiveForm := Self;
  if Assigned(FOnActivate) then FOnActivate(Sender);
end;

procedure TDpiForm.DeactivateHandler(Sender: TObject);
begin
  DpiScreen.FActiveForm := DpiScreen.FActiveFormPrev;
  //DpiScreen.UpdateActiveFormFromNativeScreen;
  if Assigned(FOnDeactivate) then FOnDeactivate(Sender);
end;

procedure TDpiForm.DoClose(var CloseAction: TCloseAction);
begin
  if Assigned(FOnClose) then FOnClose(Self, CloseAction);
end;

procedure TDpiForm.CloseHandler(Sender: TObject; var CloseAction: TCloseAction);
begin
  Close;
end;

procedure TDpiForm.CloseQueryHandler(Sender : TObject; var CanClose : boolean);
begin
  if Assigned(FOnCloseQuery) then FOnCloseQuery(Self, CanClose);
end;

procedure TDpiForm.CreateParams(var p: TCreateParams);
begin
  // TODO: NativeForm.CreateParams(P);
end;

// This method is called by TWriter to retrieve the child components to write
procedure TDpiForm.GetChildren(Proc: TGetChildProc; Root: TComponent);
var
  I: Integer;
  OwnedComponent: TComponent;
begin
  DebugLn(['TDpiForm.GetChildren ComponentCount=', ComponentCount]);
  inherited GetChildren(Proc, Root);
  if Root = Self then begin
    for I := 0 to ComponentCount - 1 do begin
      OwnedComponent := Components[I];
      if not OwnedComponent.HasParent then Proc(OwnedComponent);
    end;
  end;
end;

function TDpiForm.GetNativeScrollingWinControl: TScrollingWinControl;
begin
  Result := GetNativeForm;
end;

function TDpiForm.GetNativeForm: TForm;
begin
  if not Assigned(NativeForm) then begin
    NativeForm := TFormEx.CreateNew(nil);
    (NativeForm as TFormEx).OnMessage := @FormMessageHandler;
    //NativeForm := TForm.Create(nil);
  end;
  Result := NativeForm;
end;

procedure TDpiForm.UpdateNativeControl;
begin
  inherited;
  GetNativeForm.OnActivate := @ActivateHandler;
  GetNativeForm.OnDeactivate := @DeactivateHandler;
  GetNativeForm.OnClose := @CloseHandler;
  GetNativeForm.OnCloseQuery := @CloseQueryHandler;
  GetNativeForm.Name := Name + 'Native';
end;

procedure TDpiForm.AfterConstruction;
begin
  inherited;
  DoOnCreate;
end;

function TDpiForm.ShowModal: Integer;
begin
  Result := GetNativeForm.ShowModal;
end;

procedure TDpiForm.Close;
var
  CloseAction: TCloseAction;
  IsMainForm: Boolean;
begin
  if fsModal in FormState then
    ModalResult := mrCancel
  else
  begin
    if CloseQuery then
    begin
      // IsMainForm flag set if we are closing MainForm or its parent
      IsMainForm := (DpiApplication.MainForm = Self) or (Self.IsParentOf(DpiApplication.MainForm));
      // Prepare default close action
      if FormStyle = fsMDIChild then
      begin
        CloseAction := caNone;
        // TODO: mdi logic
      end
      else
      begin
        if IsMainForm then
          CloseAction := caFree
        else
          CloseAction := caHide;
      end;
      // call event handler and let user modify CloseAction
      DoClose(CloseAction);
      // execute action according to close action
      case CloseAction of
        caHide: Hide;
        caMinimize: WindowState := wsMinimized;
        caFree:
          begin
            // if form is MainForm, then terminate the application
            // the owner of the MainForm is the application,
            // so the Application will take care of free-ing the form
            // and Release is not necessary
            if IsMainForm then
              DpiApplication.Terminate
            else
              Release;
          end;
      end;
    end;
  end;
end;

function TDpiForm.CloseQuery: boolean;
begin
  Result := True;
  if Assigned(FOnCloseQuery) then
    FOnCloseQuery(Self, Result);
end;

procedure TDpiForm.BringToFront;
begin
  GetNativeForm.BringToFront;
end;

procedure TDpiForm.Release;
begin
  Free;
end;

// Init the component with an IDE resource
constructor TDpiForm.Create(TheOwner: TComponent);
begin
  //inherited;
  //DebugLn(['TDpiForm.Create ', DbgSName(TheOwner)]);
  GlobalNameSpace.BeginWrite;
  try
    CreateNew(TheOwner, 1); // this calls BeginFormUpdate, which is ended in AfterConstruction
    if (ClassType <> TDpiForm) and not (csDesigning in ComponentState) then begin
      if not InitResourceComponent(Self, TDataModule) then begin
        raise EResNotFound.Create('Resource missing for class ' + ClassName);
      end;
    end;
  finally
    GlobalNameSpace.EndWrite;
  end;
  ScreenChanged;
  UpdateNativeControl;
end;

constructor TDpiForm.CreateNew(AOwner: TComponent; Num: Integer);
begin
  inherited Create(AOwner);
  DpiScreen.AddForm(Self);
end;

destructor TDpiForm.Destroy;
begin
  // TODO: Can't destroy directly?
  TFormEx(NativeForm).OnMessage := nil;
  FreeAndNil(NativeForm);

  DpiScreen.RemoveForm(Self);
  inherited;
end;

initialization

RegisterPropertyToSkip(TDpiForm, 'OldCreateOrder', 'Native compatibility property', '');
RegisterPropertyToSkip(TDpiForm, 'TextHeight', 'Native compatibility property', '');
RegisterPropertyToSkip(TDpiForm, 'Scaled', 'Native compatibility property', '');
RegisterPropertyToSkip(TDpiForm, 'TransparentColorValue', 'Native compatibility property', '');
DpiScreen := TDpiScreen.Create;
DpiApplication := TDpiApplication.Create(nil);
DpiMouse := TDpiMouse.Create;
RegisterClasses([TDpiButton, TDpiImage, TDpiPaintBox, TDpiListBox]);

finalization

FreeAndNil(DpiApplication);
FreeAndNil(DpiScreen);
FreeAndNil(DpiMouse);

end.

